from typing import Union, List

import numpy as np
from pyspark import Broadcast
from pyspark.ml.linalg import DenseVector
from sklearn.base import BaseEstimator


def as_model(model: Union[BaseEstimator, Broadcast]) -> BaseEstimator:
    """Unpack model if it has been previously wrapped in a Broadcast object.

    Parameters
    ----------
    model: The model (or broadcast wrapper containing it) to be returned.

    Returns
    -------
    A scikit-learn estimator.
    """
    return (model.value
            if isinstance(model, Broadcast)
            else model)


def as_features(x: Union[DenseVector, List[float], np.ndarray]):
    """Convert a non-normalized input in a sklearn-friendly feature array.

    Parameters
    ----------
    x: The non-conforming data source representation.

    Returns
    -------
    Numpy array of exactly one sample, multiple features.
    """
    if isinstance(x, DenseVector):
        x = x.toArray()

    if not isinstance(x, np.ndarray):
        x = np.asarray(x)

    return x.reshape(1, -1)



import os
import logging
from argparse import ArgumentParser

import infra.core.forge.processors.core as C
import infra.core.forge as ARCHETYPE_INITIAL


def parse_args():
    p = ArgumentParser('NAME_SIMPLE job')
    p.add_argument('--inputs', default=os.path.join(ARCHETYPE_INITIAL.config.lakes.transient, 'NAME.parquet'))
    p.add_argument('--outputs', default=os.path.join(ARCHETYPE_INITIAL.config.lakes.refined, 'NAME.parquet'))

    return p.parse_args()


def run(inputs, outputs):
    raw = os.path.join(ARCHETYPE_INITIAL.config.lakes.raw, 'NAME.parquet')
    tru = os.path.join(ARCHETYPE_INITIAL.config.lakes.trusted, 'NAME.parquet')

    (C.Job(
        ARCHETYPE_INITIAL.processors.NAME_DOTTED.Rawing(inputs, raw),
        ARCHETYPE_INITIAL.processors.NAME_DOTTED.Trusting(raw, tru),
        ARCHETYPE_INITIAL.processors.NAME_DOTTED.Refining(tru, outputs))
     .setup(ARCHETYPE_INITIAL.config)
     .perform()
     .describe()
     .teardown())


if __name__ == '__main__':
    logging.basicConfig(**ARCHETYPE_INITIAL.config.logging.default.asDict())
    args = parse_args()
    
    run(args.inputs, args.outputs)



import abc
import logging
import os
from typing import Dict, Optional

from . import consts

C = Dict[str, str]


def default_env_path(target: str, env: str) -> str:
    return os.path.join(target, 'config', env)


def default_env_file(target: str, env: str) -> str:
    return os.path.join(default_env_path(target, env), '.env')


def load_config(env_file: str) -> C:
    if not os.path.exists(env_file):
        raise RuntimeError(f'Cannot load env file {env_file}.\nMake sure '
                           f'your current directory is a valid project.')

    with open(env_file) as f:
        ls = f.readlines()

    ls = [l.strip() for l in ls]
    ls = [l for l in ls if l and not l.startswith('#')]
    parts = [l.split('=') for l in ls]
    config = {k: '='.join(v) for k, *v in parts}

    return config


def assert_is_core_project(target):
    if not os.path.exists(os.path.join(target, 'myra')):
        raise RuntimeError('DnA project not found.\nMake sure your '
                           'current directory is a valid project.')


def context(env) -> C:
    """Retrieve the project's current context, based on an environment.

    """
    target = os.getcwd()

    env_path = default_env_path(target, env)
    env_file = default_env_file(target, env)

    c = load_config(env_file)
    c.update(_ENV_PATH=env_path,
             _ENV_FILE=env_file,
             _TARGET=target,
             _CORE_REPO=consts.CORE_REPO,
             _ARCH_REPO=consts.ARCH_REPO)

    return c


class Runner(metaclass=abc.ABCMeta):
    def __init__(self, config: C):
        self.config = config

    def execute(self,
                command: str,
                log_contextualized: bool = False,
                **context):
        command_c = self.contextualized(command, context)
        logging.info(f'Running `{command_c if log_contextualized else command}`')

        if command is NotImplemented:
            raise NotImplementedError(f'Command was not implemented for '
                                      f'`{type(self).__name__}` backend.')

        return os.system(command_c)

    def contextualized(self, command, context=None):
        c = {**self.config, **(context or {})}
        return command.format(**c)


# region Artifacts related tasks

def safe_name(name):
    return name.replace('-', '_')


def as_path(name: str,
            directory: str,
            suffix: Optional[str] = '.py'):
    name = safe_name(name)
    locations = (directory, f'./{directory}')

    path, file = os.path.split(name)
    return os.path.join(
        path if path.startswith(locations) else os.path.join(directory, path),
        file if not suffix or file.endswith(suffix) else file + suffix)

# endregion



import operator
from typing import List, Union, Any, Callable

from pyspark.sql import DataFrame, functions as F, types as T, Window, Column

from infra.core.forge import utils as U


## region column identifiers
def is_binary(df: DataFrame, column: str) -> bool:
    """Checks whether a given column in a dataframe is binary.

    Parameters
    ----------
    df : DataFrame
        The dataframe.
    column : str
        Column name.

    Return
    ------
    bool
        True if ``column`` has two possible values, False otherwise.

    """
    return isinstance(df.schema[column].dataType, T.BooleanType)


def is_categorical(df: DataFrame,
                   column: str,
                   cat_threshold: Union[int, float, None] = .9) -> bool:
    """Checks whether a given column in a dataframe is categorical.

    See Also
    --------
    is_binary   : Checks whether a given column in a dataframe is binary.
    is_discrete : Check whether a given column is either categorical or binary.

    Parameters
    ----------
    df : DataFrame
        The dataframe.
    column : str
        Column name.
    cat_threshold : int or float or None, optional
        Maximum proportion between unique values and total that is considered to be categorical. Default is .9, but
        this should be set to whatever value is sensible for your dataset.

        Values between 0 and 1 will be considered proportions, while integer values greater than 2 will be considered
        absolute quantities. Absolute value 1 is not allowed, as a single class is not useful for any calculations.

        If set to None, any checks regarding unique values proportion are ignored. This is not recommended unless the
        dataset is either very small or the column is known to be probably categorical.

        This parameter aims to avoid using id columns as categorical and end up doing expensive calculations (e.g.
        correlation) on too many classes.

    Returns
    -------
    bool
        True if ``column`` is categorical, False otherwise.

    """
    if df.schema[column].dataType != T.StringType():
        return False
    if cat_threshold is None:
        return True

    c, d = df.select(F.count(column), F.countDistinct(column)).collect()[0]

    if cat_threshold > 1:
        if not isinstance(cat_threshold, int):
            raise ValueError('`cat_threshold` should be a float between 0 and 1 representing the '
                             'distinct percentage or integer representing the count of distinct values.')
        return d <= cat_threshold

    coefficient = d / c
    return coefficient <= cat_threshold


def is_numeric(df: DataFrame, column: str):
    """Checks whether a given column in a dataframe is numeric.

    Parameters
    ----------
    df : DataFrame
        The dataframe.
    column : str
        Column name.

    Returns
    -------
    bool
        True if ``column`` is of some numeric type, False otherwise.

    """
    return isinstance(df.schema[column].dataType, T.NumericType)


def is_class(df: DataFrame, column: str, cat_threshold: Union[int, float, None] = .9):
    """Checks whether a given column in a dataframe is a classification (a category or some binary value).

    Parameters
    ----------
    df : DataFrame
        The dataframe.
    column : str
        Column name.
    cat_threshold : int or float or None, optional
        Maximum proportion between unique values and total that is considered to be categorical. Default is .9.

    See Also
    --------
    is_categorical : for a detailed explanation on ``cat_threshold``.

    Returns
    -------
    bool
        True if ``column`` is either categorical or binary, False otherwise.

    """
    return is_categorical(df, column, cat_threshold) or is_binary(df, column)


def is_correlatable(df: DataFrame,
                    column: str,
                    cat_threshold: Union[int, float, None] = .9):
    """Checks whether a given column in a dataframe can be used to calculate correlation.

    Parameters
    ----------
    df : DataFrame
        The dataframe.
    column : str
        Column name.
    cat_threshold : int or float or None, optional
        Maximum proportion between unique values and total that is considered to be categorical. Default is .9.

    See Also
    --------
    is_categorical : for a detailed explanation on ``cat_threshold``.

    Returns
    -------
    bool
        True if ``column`` is either numeric, categorical or binary. False otherwise.

    """
    return (is_binary(df, column) or
            is_numeric(df, column) or
            is_categorical(df, column, cat_threshold))


def correlatable_columns(df: DataFrame,
                         subset: Union[str, List[str]] = 'correlatable',
                         cat_threshold: Union[int, float, None] = .9):
    """Selects a subset of correlatable columns in a DataFrame.

    Parameters
    ----------
    df : DataFrame
        The dataframe.
    subset : str or list of str, optional
        Can be either a list of column names or one of:

        - ``'correlatable'``: Select all correlatable columns (default)
        - ``'binary'``: Retrieve only binary columns
        - ``'categorical'``: Retrieve only categorical (nominal) columns
        - ``'class'``: Retrieve categorical and binary columns
        - ``'numeric'``: Retrieve only numeric columns
    cat_threshold : int or float or None, optional
        Maximum proportion between unique values and total that is considered to be categorical. Default is .9.

    See Also
    --------
    is_binary      : Checks whether a given column in a dataframe is binary.
    is_categorical : Checks whether a given column in a dataframe is categorical. Also includes a detailed explanation
                     on``cat_threshold``.
    is_numeric     : Checks whether a given column in a dataframe is numeric.

    Returns
    -------
        list of str
            A list with the selected columns' namess
    """
    if subset in ['auto', 'correlatable']:
        subset = (c for c in df.columns if
                  is_correlatable(df, c, cat_threshold=cat_threshold))
    elif subset in ['binary', 'boolean', 'bool', 'b']:
        subset = (c for c in df.columns if is_binary(df, c))
    elif subset in ['categorical', 'cat', 'c']:
        subset = (c for c in df.columns if
                  is_categorical(df, c, cat_threshold=cat_threshold))
    elif subset == 'class':
        subset = (c for c in df.columns if is_class(df, c, cat_threshold=cat_threshold))
    elif subset in ['numeric', 'num', 'n']:
        subset = (c for c in df.columns if is_numeric(df, c))
    else:
        subset = U.to_list(subset)
    return set(subset)


# end region

## region sequences

def sessionize(df: DataFrame,
               threshold: Any,
               id_col: Union[str, Column] = 'id',
               ordering_col: Union[str, Column] = 'created_at',
               interval_function: Callable[[Column, Column], Any] = operator.sub,
               output_col: str = 'session',
               interval_col: str = None,
               rank_col: str = None) -> DataFrame:
    """Determines sessions for an user event dataframe.

    Parameters
    ----------
    df : pyspark.sql.DataFrame
        The dataframe.
    threshold : Any
        Maximum interval between events such that they are considered to be in the same session.
    id_col : pyspark.sql.Column or str, optional
        Name of the user identifying column, used to group events from a user into a single row. (Default: id)
    ordering_col : pyspark.sql.Column or str, optional
        Name of the column used to order the events. (Default: created_at)
    interval_function : callable, optional
        Lambda function used to evaluate intervals between events. This should receive two ``pyspark.sql.Column`` and
        return an object comparable with ``session_thresold``. Defaults to ``operator.sub``, that is, usual subtraction.
    output_col : str, optional
        Name of the output column (default: ``session``).
    rank_col, interval_column : str, optional
        Names of the columns to receive event rank (number) within the session and interval between events. None to
        ignore (default).

    Returns
    -------
    pyspark.sql.DataFrame
        The given dataframe with an additional column containing session information.

    Examples
    --------
    .. jupyter-execute:: /examples/analysis/utils/sessionize.py

    """
    window = Window.partitionBy(id_col).orderBy(ordering_col)
    df = df.withColumn('_lag', interval_function(F.col(ordering_col),
                                                 F.lag(ordering_col).over(window)))
    df = df.withColumn('_session_change',
                       F.when(F.col('_lag') > F.lit(threshold), 1).otherwise(0))
    if interval_col:
        df = df.withColumnRenamed('_lag', interval_col)
    else:
        df = df.drop('_lag')
    df = df.withColumn(output_col, F.sum('_session_change').over(window))
    if rank_col:
        window = window.partitionBy(id_col, output_col)
        df = df.withColumn(rank_col, F.rank().over(window))
    return df.drop('_session_change')


def sequentiate(df: DataFrame,
                id_col: Union[str, Column] = 'id',
                ordering_col: Union[str, Column] = 'created_at',
                event_col: Union[str, Column] = 'event',
                session_col: Union[str, Column] = None,
                output_col: str = 'journey',
                partial_output_col: str = None,
                rank_output_col: str = None,
                max_depth: int = None) -> DataFrame:
    """Determines the sequence of events for each user during each session given a user event dataframe.

    Parameters
    ----------
    df : pyspark.sql.DataFrame
        The dataframe.
    id_col : pyspark.sql.Column or str, optional
        User identifying column, used to group events from a user into a single row. (Default: id)
    ordering_col : pyspark.sql.Column or str, optional
        Column used to order the events. (Default: created_at)
    event_col : pyspark.sql.Column or str, optional
        Column with the event information. (Default: event)
    session_col : str, optional
        Name of the column with the session identifier. Set to ``None`` for no session grouping (default).
    output_col : str, optional
        Name of the output column (default: ``journey``).
    partial_output_col : str, optional
        Name of the column for partial output (default: None).
        Partial output for this function is the sequence of event up until (and including) the current event.
    rank_output_col: str, optional
        Name of the column for rank output (default: None).
        Rank is the position of some event on the full sequence of events.
    max_depth : int, optional
        Maximum event sequence length. None for unlimited (default).

    See Also examples/
    --------
    sessionize : Determine sessions for an user event dataframe.

    Returns
    -------
    pyspark.sql.DataFrame
        The given dataframe with an additional column containing event sequence information.

    Examples
    --------
    .. jupyter-execute:: /examples/analysis/utils/journey.py

    """
    if session_col:
        window = Window.partitionBy(id_col, session_col)
    else:
        window = Window.partitionBy(id_col)
    window = window.orderBy(ordering_col)

    df = df.withColumn('_rank', F.rank().over(window))
    if max_depth:
        df = df.filter(F.col('_rank') <= max_depth)

    if partial_output_col:
        df = df.withColumn(output_col, F.collect_list(event_col).over(window))

    if output_col:
        df = df.withColumn(output_col, F.collect_list(event_col).over(
            window.rowsBetween(Window.unboundedPreceding, Window.unboundedFollowing)))

    if rank_output_col:
        df = df.withColumnRenamed("_rank", rank_output_col)
    else:
        df = df.drop("_rank")

    identity = [id_col, session_col] if session_col else [id_col]
    df = df.drop_duplicates(identity)

    return df

# end region



from pyspark.sql import functions as F


def anonymize(col: F.Column,
              token: str = '',
              bits: int = 256,
              length: int = None) -> F.Column:
    col = F.col(col).cast('string')
    col = F.concat(col, F.lit(token)) if token else col
    col = F.sha2(col, bits)
    col = F.substring(col, 0, length) if length else col

    return col



import logging
import time
from datetime import datetime
from typing import Any, List, Optional, Union, Dict


def is_list(x: Any) -> bool:
    """Check if input is a list or list-like.

    Parameters
    ----------
    x: any
       the object being checked, usually a frame/list of
       frames or record in the processing pipeline

    Returns
    -------
    bool
        Whether the input is a list-like object or not.

    """
    return isinstance(x, (list, tuple))


def to_list(x: Any) -> List[Any]:
    """Ensure the parameter is a list by wrapping it if necessary.

    Parameters
    ----------
    x: any
       the entry being evaluated, usually a frame
       or data record in the processing stream

    Returns
    -------
    list
        The input :code:`x` if it is a list. Otherwise, :code:`[x]` is passed.

    """
    return x if is_list(x) else [x]


def unpack(x: List[Any], on_many: str = 'warn') -> Any:
    """Undoes the :func:`to_list` function by extracting the first element.

    If the list has more than one element, :code:`on_many` parameter will
    dictate the behavior of this function.

    Parameters
    ----------
    x: list
        the list to be unpacked
    on_many: str
        what to do whenever the input list :code:`x` contains more
        than one element. "warn" is the default, the first element is unpacked
        and a warning is signaled stating about possible data loss. If "keep"
        is passed, unpacking is skipped

    Returns
    -------
    any
        The input unwrapped.

    """
    if not is_list(x):
        return x

    if len(x) <= 1:
        return next(iter(x))

    if on_many == 'warn':
        logging.warning(f'Multiple values to unpack from {x}. '
                        f'Some information may be lost.')

        return next(iter(x))

    if on_many == 'keep':
        return x

    raise ValueError(f'on_many parameter "{on_many}" is illegal for unpack function. '
                     f'Try "warn" to unpack the first element regardless the input size '
                     f'or "keep" to prevent unpacking if list has more than one element.')


def valid_or_default(a: Optional[Any], b: Any) -> Any:
    """Returns the first parameter a if it's valid.

    Parameters
    ----------
    a: any
       the parameter checked
    b: any
       the second choice in case :code:`a` is invalid

    Returns
    -------
    any
        :code:`a` if it is valid. :code:`b` otherwise.

    """
    return a if a is not None else b


def as_row(obj: Union[Dict, List[Dict]]):
    """Convert a dict into a :code:`pyspark.sql.Row`.

    If a list of dicts is passed, then this function is recursively
    applied to each element of the list.

    Parameters
    ----------
    obj: dict or list of dictionaries
         the object that will be converted into a Row

    Returns
    -------
    :class:`pyspark.sql.Row`
        A Row object containing all of the information within
        :code:`obj`.

    """
    from pyspark.sql import Row

    if isinstance(obj, dict):
        d = {k: as_row(v) for k, v in obj.items()}
        return Row(**d)

    if is_list(obj):
        return [as_row(v) for v in obj]

    return obj


class NamedEntityMixin:
    @classmethod
    def fullname(cls) -> str:
        """Retrieve a fully-representative name for this entity.
        """
        m = cls.__module__
        if m is None or m == str.__class__.__module__ or str(m) == '__main__':
            return cls.__name__  # Avoid reporting __builtin__
        return m + '.' + cls.__name__


class stopwatch:
    """Measure elapsed time for python-operations.

    This will consider blocked wall time as well, as much of the operations
    are committed during :code:`spark.DataFrame#write`.

    Parameters
    ----------
    label: str
        title used in report

    mode: str
        type of report produced.
        Options are 'short', 'long' or 'silent'

    Examples
    --------
    .. jupyter-execute::

        import dextra.dna.core as C

        with C.utils.stopwatch('Loading iris'):
            print('Samples loaded:', C.datasets.iris().count())

    """

    def __init__(self, label: str = 'execution time', mode: str = 'long'):
        if mode not in ('short', 'long', 'silent'):
            raise ValueError(f'report parameter ({mode}) should be '
                             f'either "short", "long" or "silent".')

        self.label = label
        self.mode = mode

        self.started_at = None
        self.ended_at = None
        self.perf_start = None

    @property
    def elapsed(self):
        """Time, in seconds, between start and end of the window
           profiled by the :code:`stopwatch`.
        """
        return self.perf_end - self.perf_start

    @property
    def _indent(self):
        return ' ' * (len(self.label) - len(self.label.lstrip()) + 2)

    def __enter__(self):
        self.started_at = datetime.now()
        self.perf_start = time.perf_counter()

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.ended_at = datetime.now()
        self.perf_end = time.perf_counter()

        if self.mode == 'short':
            print(f'{self.label}: {self.elapsed:.3f} seconds')
        elif self.mode == 'long':
            print(self.label)
            print(f'{self._indent}started at: {self.started_at}')
            print(f'{self._indent}ended at:   {self.ended_at}')
            print(f'{self._indent}elapsed:    {self.elapsed:.3f} seconds')



import abc
import logging
import os

import yaml

from . import consts
from .base import Runner, assert_is_core_project, context, as_path


class Backend(Runner, metaclass=abc.ABCMeta):
    """Contains basic interaction logic with a environment backend.

    This is a base-class. Look at its sub-classes for implementation details.
    """

    artifacts = {a: NotImplemented for a in consts.ARTIFACTS}

    transport = {
        'local': NotImplemented,
        'gcloud': NotImplemented
    }

    class Command:
        SYNC: str = '{TRANSPORT} {ARTIFACT_A} {ARTIFACT_B}'
        BUILD: str = NotImplemented
        CLEAN: str = NotImplemented
        START: str = NotImplemented
        EXPLORE: str = NotImplemented
        STOP: str = NotImplemented
        RUN: str = NotImplemented
        TEST: str = NotImplemented

    def build(self):
        assert_is_core_project(self.config['_TARGET'])
        self.build_dependencies()
        return self.execute(self.Command.BUILD)

    def clean(self):
        return self.execute(self.Command.CLEAN)

    def start(self):
        return self.execute(self.Command.START)

    def explore(self):
        return self.execute(self.Command.EXPLORE)

    def stop(self):
        return self.execute(self.Command.STOP)

    def run(self, job, job_args):
        assert_is_core_project(self.config['_TARGET'])
        return self.execute(self.Command.RUN,
                            JOB=as_path(job, 'jobs'),
                            JOB_ARGS=' '.join(job_args))

    def test(self, job_args):
        assert_is_core_project(self.config['_TARGET'])
        return self.execute(self.Command.TEST, JOB_ARGS=' '.join(job_args))

    def sync(self, artifact, target):
        assert_is_core_project(self.config['_TARGET'])
        t = adapter(target, verbose=False)

        return self.execute(self.Command.SYNC,
                            TRANSPORT=self.transport[t.config['BACKEND']],
                            ARTIFACT_A=self.contextualized(self.artifacts[artifact]),
                            ARTIFACT_B=t.contextualized(t.artifacts[artifact]),
                            log_contextualized=True)

    def build_dependencies(self):
        for dep in consts.DEPS:
            name, repo, v_field = dep

            if v_field in self.config:
                version = self.config[v_field]

                wheel = self.contextualized('{_TARGET}/lib/wheels/%s-%s-py3-none-any.whl' % (name, version))
                source = self.contextualized('{_TARGET}/lib/wheels/%s_%s' % (name, version))

                if not os.path.exists(wheel):
                    self.clone_and_build_py_dep(
                        repo=repo,
                        branch=version,
                        source_dst=source)

    def clone_and_build_py_dep(self, repo, branch, source_dst):
        self.execute(
            f"""
    rm -rf {source_dst}
    mkdir -p {source_dst}
    git -c advice.detachedHead=false clone -b {branch} {repo} -q --depth 1 {source_dst}
    cd {source_dst}
    python setup.py -q sdist bdist_wheel --dist-dir ../
    cd ../../../..
    rm -rf {source_dst}""")


class Local(Backend):
    artifacts = {
        'notebooks': 'notebooks',
        'jobs': 'jobs'
    }

    transport = {
        'local': 'rsync -ru',
        'gcloud': 'gsutil -mq rsync -ru'
    }

    class Command(Backend.Command):
        BUILD = 'docker-compose --env-file {_ENV_FILE} build'
        CLEAN = 'rm -rf {_TARGET}/lib/wheels'
        START = 'docker-compose --env-file {_ENV_FILE} up -d'
        STOP = 'docker-compose --env-file {_ENV_FILE} down'
        RUN = 'docker-compose --env-file {_ENV_FILE} exec {SERVICE} python {JOB} {JOB_ARGS}'
        TEST = 'docker-compose --env-file {_ENV_FILE} run --rm {SERVICE} tests {JOB_ARGS}'
        EXPLORE = 'firefox http://localhost:{JUPYTER_PORT}'


class GCloud(Backend):
    artifacts = {
        'notebooks': 'gs://{BUCKET}/notebooks/jupyter',
        'jobs': 'gs://{BUCKET}/jobs',
    }

    transport = {
        'local': 'gsutil -mq rsync -ru',
        'gcloud': 'gsutil -mq rsync -ru'
    }

    class Command(Backend.Command):
        START = 'gcloud --project {PROJECT} dataproc clusters import {CLUSTER_NAME} --source {_ENV_PATH}/{CLUSTER_DEF} --region {REGION}'
        STOP = 'gcloud --project {PROJECT} dataproc clusters delete {CLUSTER_NAME} -q --region {REGION}'
        RUN = 'gcloud --project {PROJECT} dataproc jobs submit pyspark {JOB} --cluster {CLUSTER_NAME} --region {REGION} -- {JOB_ARGS}'
        EXPLORE = 'firefox $(gcloud --project {PROJECT} dataproc clusters describe {CLUSTER_NAME} --region {REGION} | grep -Pom1 "Jupyter: \K.+")'
        CLEAN = 'gsutil -mq rm -rf {_TARGET}/lib/wheels'
        BUILD = """
  python setup.py -q sdist bdist_wheel --dist-dir {_TARGET}/lib/wheels/
  gsutil mb -p {PROJECT} -l {REGION} gs://{BUCKET}
  gsutil -mq rsync -rd config/{ENV} gs://{BUCKET}/config/
  gsutil -mq cp -r lib/* gs://{BUCKET}/config/
  rm -rf build dist
  """


BACKENDS = {
    'local': Local,
    'gcloud': GCloud
}


def adapter(env, verbose: bool = True) -> Backend:
    """Retrieve the appropriate adapter for an environment.

    """
    global BACKENDS

    if verbose: logging.info(f'Using {env} environment.')

    c = context(env)
    k = c.get('BACKEND', 'local')

    if verbose: logging.debug(yaml.dump({'config': c}))

    try:
        _cls = BACKENDS[k] if isinstance(k, str) else k
        return _cls(c)

    except KeyError:
        raise RuntimeError(f'Sorry. We are not sure how to build an environment '
                           f'within `{k}` backend.\nMake sure a valid value for BACKEND '
                           f'{list(BACKENDS.keys())} is set at "config/{env}/.env"')



import io
import os
from typing import Tuple, List, Optional, Union, BinaryIO

import google
from google.auth.transport.requests import AuthorizedSession
from google.cloud import storage as gcs
from google.resumable_media.requests import ResumableUpload

from .file_system import FileSystemStorageAdapter
from .storage_adapter import StorageAdapter


class IntermediateBuffer(io.BufferedIOBase, BinaryIO):
    def __init__(self):
        self._buffer = b''
        self._offset = 0
        self._length = 0

    def write(self, data: Union[bytes, bytearray]) -> int:
        size = len(data)
        self._buffer += data
        self._length += size
        del data
        return size

    def read1(self, size: Optional[int] = -1) -> bytes:
        if size == -1:
            return self.readall()
        self._offset += min(self.size - self._offset, size)
        data, self._buffer = self._buffer[:size], self._buffer[size:]
        return data

    def readall(self):
        self._offset = self._length
        return self._buffer

    read = read1

    def tell(self):
        return self._offset

    def __len__(self):
        return self._length

    size = property(__len__)

    def seekable(self) -> bool:
        return False

    def writable(self) -> bool:
        return True

    def readable(self) -> bool:
        return True


class GoogleCloudStorageIO(io.BufferedIOBase, BinaryIO):
    BUFFER_SIZE = 8 * 1024 * 1024

    def __init__(self,
                 blob: gcs.Blob,
                 resumable_upload: ResumableUpload = None,
                 output_buffer: IntermediateBuffer = None):
        super().__init__()
        self._blob = blob
        self._offset = 0

        self._resumable_upload = resumable_upload
        self._output_buffer = output_buffer
        self._transport = None
        self._output_buffer_bytes = 0

        self._input_buffer = None

    @property
    def size(self):
        if self._blob.size is None:
            self._blob.reload()
        return self._blob.size

    @property
    def _buffered_bytes(self):
        return self._input_buffer.size - self._input_buffer.tell()

    @property
    def _ended(self):
        return self._offset >= self.size - 1

    def _feed_buffer(self, size: int):
        self._input_buffer = self._input_buffer or IntermediateBuffer()
        while self._buffered_bytes < size and not self._ended:
            end = min(self._offset + max(self.BUFFER_SIZE, size), self.size) - 1
            self._blob.download_to_file(self._input_buffer,
                                        start=self._offset,
                                        end=end)
            self._offset = end

    def read1(self, size: Optional[int] = -1) -> bytes:
        try:
            if size == -1:
                return self.readall()
            else:
                if self._offset >= self.size or size == 0:
                    return b''
                self._feed_buffer(size)
                data = self._input_buffer.read(size)
                return data
        except google.api_core.exceptions.NotFound as error:
            raise FileNotFoundError(error)

    read = read1

    def readinto(self, b: bytearray) -> Optional[int]:
        try:
            if self._offset >= self.size:
                return 0
            with io.BytesIO(b) as stream:
                start = stream.tell()
                self._blob.download_to_file(stream, start=self._offset)
                self._offset = self.size
                return stream.tell() - start
        except google.api_core.exceptions.NotFound as error:
            raise FileNotFoundError(error)

    def readall(self) -> bytes:
        try:
            if self._offset >= self.size:
                return b''
            with io.BytesIO() as stream:
                self._blob.download_to_file(stream, start=self._offset)
                self._offset = self.size
                return stream.getvalue()
        except google.api_core.exceptions.NotFound as error:
            raise FileNotFoundError(error)

    def _initiate_upload(self):
        if self._resumable_upload is None:
            self._resumable_upload = ResumableUpload(
                upload_url=f'https://www.googleapis.com/upload/storage/v1/b/'
                        f'{self._blob.bucket.name}/o?uploadType=resumable',
                chunk_size=self.BUFFER_SIZE)
            self._transport = AuthorizedSession(credentials=self._blob.client._credentials)
        
            self._output_buffer = self._output_buffer or IntermediateBuffer()
            self._resumable_upload.initiate(self._transport,
                                            stream=self._output_buffer,
                                            metadata={'name': self._blob.name},
                                            content_type='application/octet-stream',
                                            stream_final=False)

    def _transmit_next_chunk(self):
        try:
            self._resumable_upload.transmit_next_chunk(self._transport)
        except google.resumable_media.common.InvalidResponse:
            self._resumable_upload.recover(self._transport)
        self._output_buffer_bytes -= min(self._output_buffer_bytes, self.BUFFER_SIZE)

    def write(self, b: Union[bytes, bytearray, BinaryIO]) -> int:
        self._initiate_upload()
        if isinstance(b, (bytes, bytearray)):
            self._output_buffer.write(b)
            self._output_buffer_bytes += len(b)
            while self._output_buffer_bytes >= self.BUFFER_SIZE:
                self._transmit_next_chunk()
        else:
            for chunk in iter(lambda: b.read(self.BUFFER_SIZE), b''):
                self.write(chunk)

    def flush(self):
        while self._output_buffer_bytes > 0:
            self._transmit_next_chunk()

    def seek(self, offset: int, whence: int = io.SEEK_SET) -> int:
        if whence == io.SEEK_SET:
            self._offset = offset
        elif whence == io.SEEK_CUR:
            self._offset += offset
        elif whence == io.SEEK_END:
            self._offset = self.size + offset
        else:
            raise ValueError("Invalid value for `whence`")
        return self._offset

    def tell(self) -> int:
        return self._offset

    def readable(self) -> bool:
        return True

    def writable(self) -> bool:
        return True

    def seekable(self) -> bool:
        return True

    def close(self) -> None:
        if self._output_buffer is not None:
            self.flush()
            del self._output_buffer
        if self._input_buffer is not None:
            del self._input_buffer
        super().close()


class GoogleCloudStorageAdapter(StorageAdapter):
    def __init__(self, client: Optional[gcs.Client] = None):
        super().__init__()
        self._client = client

    def duplex(self, filename: str) -> GoogleCloudStorageIO:
        return GoogleCloudStorageIO(self._blob(filename))

    reader = duplex

    def writer(self, filename: str, append: bool = False) -> GoogleCloudStorageIO:
        self.makedirs(filename)
        stream = self.duplex(filename)
        if append:
            stream.seek(0, os.SEEK_END)
        return stream

    @staticmethod
    def _split_path(filename: str) -> Tuple[str, str]:
        bucket, *file = filename.split('/')
        return bucket, '/'.join(file)

    def _blob(self, filename: str) -> gcs.Blob:
        bucket, file = self._split_path(filename)
        bucket = self.client.get_bucket(bucket)
        return bucket.get_blob(file) or bucket.get_blob(file + '/') or bucket.blob(file)

    @property
    def client(self):
        if self._client is None:
            self._client = gcs.Client()
        return self._client

    def isdir(self, path: str) -> bool:
        return (self.exists(path) and
                (not self._split_path(path)[1] or  # when bucket itself
                 self._blob(path).name.endswith('/')))

    def isfile(self, path: str) -> bool:
        return self.exists(path) and not self.isdir(path)

    def _listdir(self, dirname: str) -> List[str]:
        if not dirname.endswith('/'):
            dirname += '/'

        bucket_name, path = self._split_path(dirname)
        bucket = self.client.get_bucket(bucket_name)

        iterator = bucket.list_blobs(prefix=path, fields='items/name', delimiter='/')
        iterator.extra_params['includeTrailingDelimiter'] = True
        blobs = {b.name for b in iterator if b.name != path}

        iterator = bucket.list_blobs(prefix=path, fields='prefixes', delimiter='/')
        iterator.extra_params['includeTrailingDelimiter'] = True
        iterator._items_key = 'prefixes'
        iterator.item_to_value = lambda it, x: x
        blobs.update({b for b in iterator})

        try:
            return [b[len(path):].strip('/') for b in blobs]
        except google.api_core.exceptions.NotFound as error:
            raise FileNotFoundError(error)

    def delete(self, filename: str):
        try:
            if self.isdir(filename):
                for sub in self.listdir(filename):
                    self.delete(os.path.join(filename, sub))
            self._blob(filename).delete()
        except google.api_core.exceptions.NotFound as error:
            raise FileNotFoundError(error)

    def makedirs(self, path: str):
        path_ = ''
        for p in path.split('/')[:-1]:
            path_ += p + '/'
            if not self.exists(path_):
                self._blob(path_).upload_from_string(b'')

    def copy(self, source: str, target: str, other: Optional[StorageAdapter] = None, move: bool = False):
        super()._pre_copy(source, target, other)

        if isinstance(other, FileSystemStorageAdapter):
            return self._blob(source).download_to_filename(target)

        if self.isdir(source) or not isinstance(other, GoogleCloudStorageAdapter):
            return super().copy(source, target, other, move=move)

        try:
            source_bucket, source_file = self._split_path(source)
            target_bucket, target_file = self._split_path(target)
            blob = self._blob(source)
            if move and source_bucket == target_bucket:
                self.client.get_bucket(source_bucket).rename_blob(blob, target_file)
            else:
                self.client.get_bucket(source_bucket).copy_blob(blob,
                                                                self.client.get_bucket(target_bucket),
                                                                target_file)
                if move:
                    blob.delete()
        except google.api_core.exceptions.NotFound as error:
            raise FileNotFoundError(error)

    def exists(self, path: str) -> bool:
        bucket, filename = self._split_path(path)
        if not filename:
            return self.client.bucket(bucket).exists()
        return self._blob(path).exists()



import os
from typing import Dict, Any, List, Optional

from pyspark.sql import functions as F

from .stream_adapter import StreamAdapter
from infra.core.forge.utils.configs import spark

ALLOWED_INPUT_FILE_FORMATS = ('.csv', '.json', '.parquet')
DEFAULT_READING_OPTIONS = {
    'csv': {'header': True},
    'json': {},
    'parquet': {}
}


class FileStreamAdapter(StreamAdapter):
    def read(self,
             filename: str,
             **options: Dict[str, Any]) -> F.DataFrame:
        _, ext = os.path.splitext(filename)

        if ext not in ALLOWED_INPUT_FILE_FORMATS:
            raise ValueError(f'Format {ext} is not permitted as input. '
                             f'Allowed values are: {ALLOWED_INPUT_FILE_FORMATS}')

        reader = spark().read
        options = options or DEFAULT_READING_OPTIONS[ext[1:]]

        for k, v in options.items():
            reader = reader.option(k, v)

        return reader.format(ext[1:]).load(filename)

    def write(self,
              df: F.DataFrame,
              filename: str,
              mode: str = 'overwrite',
              coalesce: Optional[int] = None,
              partitions: List[str] = None):
        if coalesce:
            df = df.coalesce(coalesce)

        writer = df.write
        writer = writer.mode(mode)

        if partitions:
            writer.partitionBy(partitions)

        writer.parquet(filename)



import importlib
import logging
from typing import Dict, Any, Union, Optional, Type, Tuple

PROTOCOL_SEP = '://'


def _assert_known_protocol(protocol, known_protocols):
    if protocol not in known_protocols:
        available = ', '.join(known_protocols)
        raise ValueError(f'Protocol scheme `{protocol}` is not supported. '
                         f'Available protocols are: {available}.')


def adapter(protocol: str,
            adapters: Dict[str, Any],
            adapters_available: Dict[str, str]):
    """Retrieve the appropriate adapter for a given protocol.

    If the adapter is not found in the :attr:`adapters` pool, its class will
    be fetched from :attr:`adapters_available`, instantiated and added to the pool.
    No adapter (nor its module) will be imported into the program's memory unless
    necessary. This allows us to clip dependencies for sources we are not using,
    as no direct references are pointed.
    Example: a project that only runs locally can safely remove the
    `google-cloud-storage` dependency and still work, as the GoogleStorageAdapter
    will never be imported.

    Parameters
    ----------
    protocol: str
        the protocol of interest
    adapters: dict of adapters started
    adapters_available: dict of paths to adapters registered

    Returns
    -------
    Adapter
        Storage or stream adapter
    """
    if protocol not in adapters:
        _assert_known_protocol(protocol, adapters_available)

        module, cls = adapters_available[protocol]
        module = importlib.import_module(module, __package__)
        cls = getattr(module, cls)
        adapters[protocol] = cls()
    return adapters[protocol]


def split_protocol_path(location: str) -> Tuple[str, str]:
    """Separates and returns protocol and path from a location.

    Parameters
    ----------
    location : str
        A valid URI with format ``[protocol://]path``. If ``protocol://`` is not present, ``file://`` is inferred.

    Returns
    -------
    protocol : str
    path : str

    """
    parts = location.split(PROTOCOL_SEP)

    if len(parts) > 2:
        raise ValueError(f'Malformed location "{location}".')

    if len(parts) == 1:
        return 'file', location

    protocol, path = parts
    return protocol, path


def without_protocol(location: str) -> str:
    """Returns a location without its associated protocol.

    Parameters
    ----------
    location : str
        A valid URI with format ``[protocol://]path``

    Returns
    -------
    str

    """
    return split_protocol_path(location)[-1]


def join_protocol_path(protocol: str, location: str) -> str:
    """Joins a protocol and some location.

    Parameters
    ----------
    protocol : str
    location : str

    Returns
    -------
    str
        A location with format ``protocol://path``, where ``path`` is the given location with its protocol removed, if
        it has any.

    """
    return f'{protocol}://{without_protocol(location)}'


# region error handling

ON_ERROR_HANDLERS = ('raise', 'log')


def handle_error(error: Union[Exception, str],
                 on_error: str = 'raise',
                 wrapper_cls: Optional[Type[Exception]] = None):
    """Raises or logs occurring errors based on the caller choice.

    :param error: the error that occurred
    :type error: Exception, str
    :param on_error: how to handle the error (raise, log or ignore)
    :type on_error: str
    :param wrapper_cls: an exception class to wrap around the error.
                        This is useful whenever you want to merge
                        errors from multiple backends into a single
                        base error.
    :type wrapper_cls: Type[Exception]
    """
    if on_error == 'raise':
        raise wrapper_cls(error) if wrapper_cls else error

    if on_error == 'log':
        return logging.error(error)

    if on_error == 'ignore':
        return

    if on_error not in ON_ERROR_HANDLERS:
        raise ValueError(f'`handle_error` function cannot understand '
                         f'on_error="{on_error}". It should assume '
                         f'one of the following values: {ON_ERROR_HANDLERS}.')

# endregion



import logging
import uuid
from typing import Callable, Optional, Tuple, List, Union

import joblib
import joblibspark
import numpy as np
from pyspark import ml
from pyspark.sql import functions as F
from sklearn import metrics
from sklearn.base import BaseEstimator, is_classifier
from sklearn.utils import parallel_backend

from .base import Driver
from .. import errors, functions
from ..base import logistic_regressor_gs
from ... import io
from infra.core.forge.utils.configs import spark, Config

from infra.core.forge.utils.utils import valid_or_default, is_list

E = Callable[[str], BaseEstimator]
FE = Union[str, List[str]]
XY = Tuple[np.ndarray, np.ndarray]

joblibspark.register_spark()


class SkLearn(Driver):
    """Scikit-learn Models Driver.

    When used as is, this driver will serve logistic regression classifiers,
    trained using GridSearch 5-2 cross-validation strategy.

    Arguments
    ---------
    config: Config
        Reference to current run's configuration.
    caching: bool
        Whether RAM-level caching should be kept.
    estimator_fn: Callable
        Building function of the serving model. By Default, uses a
        :class:`GridSearchCV` wrapped around a
        :class:`LogisticRegression`.
        (dna.core.models.logistic_regression_gs).
    scoring: str
        String representing the scoring method for the model.
        Ignored if :func:`~SkLearn.evaluate` is not called.
    broadcasting: bool
        Whether to broadcast the model across all cluster nodes.

    """

    ESTIMATOR_FN: E = staticmethod(logistic_regressor_gs)
    SCORING: str = 'balanced_accuracy'
    BROADCASTING: bool = True

    def __init__(self,
                 config: Config,
                 caching: Optional[bool] = None,
                 estimator_fn: Optional[E] = None,
                 scoring: Optional[str] = None,
                 broadcasting: Optional[bool] = None,
                 weights_prefix: str = ''):
        super().__init__(config=config, caching=caching, weights_prefix=weights_prefix)

        self.estimator_fn = valid_or_default(estimator_fn, self.ESTIMATOR_FN)
        self.scoring = valid_or_default(scoring, self.SCORING)
        self.broadcasting = valid_or_default(broadcasting, self.BROADCASTING)

    def build(self):
        logging.debug(f'driver build {self.fullname()}')
        return self.estimator_fn(scoring=self.scoring)

    def load(self, weights):
        logging.debug(f'driver load {weights}')
        return self.cache(weights) or self.load_from_storage(weights)

    def load_from_storage(self, weights: str) -> BaseEstimator:
        """Load model from the storage backend.

        Parameters
        ----------
        weights: str
            the path into which the model should be saved

        Returns
        -------
        BaseEstimator
            The loaded sklearn model

        """
        logging.debug(f'driver load model from storage {weights}')
        protocol, location = io.storage.split_protocol_path(weights)

        if protocol != 'file':
            location = f'/tmp/{uuid.uuid4()}'

            try:
                io.storage.copy(source=weights, target=location)
            except FileNotFoundError as error:
                raise errors.TrainingNotFound(error)

        try:
            model = joblib.load(location)
            self.cache(weights, model)

        except FileNotFoundError as error:
            raise errors.TrainingNotFound(error)

        return model

    def save(self,
             model: BaseEstimator,
             weights: str) -> 'SkLearn':
        """Save a scikit-learn model in the storage backend.

        This method first saves the model in a temporary location so it can
        then be copied into the running storage backend (e.g. GCS).

        Parameters
        ----------
        model: ML Model
            the model being saved
        weights: str
            the path into which the model should be saved
        """
        logging.debug(f'driver save {weights} {model}')
        self.cache(weights, model)

        location = f'/tmp/{uuid.uuid4()}'
        joblib.dump(model, location)
        io.storage.copy(source=location, target=weights)

        return self

    def learn(self,
              x: F.DataFrame,
              features: FE = 'features',
              target: str = 'target',
              weights: str = None) -> 'SkLearn':
        logging.debug(f'driver learn {weights}')

        model = self.build()

        x = self.normalize_features(x, features)
        x = x.select('features', F.col(target).alias('target'))
        x = self._balanced(x)
        x, y = self.extract_arrays(x)

        try:
            with parallel_backend('spark', n_jobs=3):
                model.fit(x, y)
        except ValueError as error:
            raise errors.Training(error)

        self.save(model, weights or self.location_from_target(target))

        return self

    def evaluate(self,
                 x: F.DataFrame,
                 features: FE = 'features',
                 target: str = 'target',
                 weights: str = None):
        weights = weights or self.location_from_target(target)
        logging.debug(f'driver evaluate {weights}')

        model = self.load(weights)

        x = self.normalize_features(x, features)
        x = x.select('features', F.col(target).alias('target'))
        z, y = self.extract_arrays(x)

        scorer = metrics.get_scorer(self.scoring)
        score = scorer(model, z, y)

        if is_classifier(model):
            logging.debug(f'classification report over all data ({len(y)} samples):')
            logging.debug(metrics.classification_report(y, model.predict(z)))

        logging.info(f'{self.scoring} score: {score}')

        return score

    def infer(self,
              x: F.DataFrame,
              features: FE = 'features',
              target: str = 'target',
              weights: str = None,
              on_errors: str = 'raise') -> F.DataFrame:
        weights = weights or self.location_from_target(target)
        logging.debug(f'driver infer {weights}')

        try:
            model = self.load(weights)

            if self.broadcasting:
                model = spark().sparkContext.broadcast(model)

            x = self.normalize_features(x, features)
            return x.withColumn(target, functions.predict('features', model))

        except errors.TrainingNotFound:
            if on_errors == 'raise':
                raise
            return x.withColumn(target, F.lit(None))

    def location_from_target(self, target: str) -> str:
        """Determine a model's location from the target column.

        We append the suffix :code:`.p` here in order to represent
        a :code:`pickle` file.

        Parameters
        ----------
        target : The target column in which the data is trained over.

        Returns
        -------
        str
            The full location of the model.
        """
        return super().location_from_target(target) + '.p'

    def normalize_features(self, x: F.DataFrame, features: FE) -> F.DataFrame:
        """Collect feature columns into a single Vector column,
        if it hasn't happened yet.

        Parameters
        ----------
        x: DataFrame
            the frame onto which the model will be applied
        features: str
            columns containing the features for the model

        Returns
        -------
        DataFrame
            A frame containing the 'features' column in it

        """
        if is_list(features):
            assembler = ml.feature.VectorAssembler(inputCols=features,
                                                   outputCol='features')
            x = assembler.transform(x)

        if isinstance(features, str) and features != 'features':
            return x.withColumn('features', F.col(features))

        return x

    def extract_arrays(self, x: F.DataFrame) -> XY:
        """Extract the DataFrame features as numpy arrays.

        Parameters
        ----------
        x: DataFrame
            the frame onto which the model will be applied

        Returns
        -------
        tuple(ndarray, ndarray)
            The numpy array pair (x, y) used during training or evaluation.
        """
        x_ = x.select('features', 'target').collect()
        z = np.asarray([r.features.toArray() for r in x_])
        y = np.asarray([d.target for d in x_])

        return z, y

    def __repr__(self):
        return (f'<{self.__module__}.{type(self).__name__} at {hex(id(self))}, '
                f'estimator fn: {self.estimator_fn.__name__}, '
                f'scoring: {self.scoring}, '
                f'config: {self.config.env}>')


class BatchedSkLearn(SkLearn):
    """Driver for Scikit-learn models that leverage vectorized user functions.
    It currently supports only structured data.
    """
    def infer(self,
              x: F.DataFrame,
              features: List[str] = None,
              target: str = 'target',
              weights: str = None,
              on_errors: str = 'raise') -> F.DataFrame:
        weights = weights or self.location_from_target(target)
        logging.debug(f'driver infer {weights}')

        self._assert_features_list(features)

        try:
            model = self.load(weights)

            if self.broadcasting:
                model = spark().sparkContext.broadcast(model)

            return x.withColumn(target, functions.predict_batch(features, model))

        except errors.TrainingNotFound:
            if on_errors == 'raise':
                raise
            return x.withColumn(target, F.lit(None))

    def normalize_features(self, x: F.DataFrame, features: FE) -> F.DataFrame:
        self._assert_features_list(features)
        return super().normalize_features(x, features)

    def _assert_features_list(self, features: FE):
        if not features or not isinstance(features, List):
            raise ValueError(
                f'`BatchedSkLearn` driver can only predict using a list of '
                f'structured features, but {features.__class__.__name__} '
                f'`{features}` was passed. Check the docs for more on the '
                f'limitations of this driver.')



import logging
from argparse import ArgumentParser

import infra.core.forge as ARCHETYPE_INITIAL


def parse_args():
    p = ArgumentParser('NAME_SIMPLE job')
    p.add_argument('name')
    
    return p.parse_args()


def run(name):
    logging.info(f'Hello {name}!')
    logging.info(f'Lakes: {ARCHETYPE_INITIAL.config.lakes}')


if __name__ == '__main__':
    logging.basicConfig(**ARCHETYPE_INITIAL.config.logging.default.asDict())
    args = parse_args()

    run(args.name)



"""Correlation for large-scale distributed data.

Contains multiple functions to infer relationship between binary, categorical
and continuous features.

"""
import warnings
from math import sqrt
from typing import Union, Tuple, Callable, List

import numpy as np
import pandas as pd
from pyspark.sql import DataFrame, functions as F
from scipy import stats

from . import utils as AU


class ChiSquareSmallFrequencyWarning(UserWarning):
    """Warned when chi-squared's result cannot be guaranteed
    because of data instability.
    """


class CategoricalCorrelationSingleClassWarning(UserWarning):
    """Warned when a feature/columns contains a single categorical value.

    In this case, aggregations cannot be produced in order to find
    significant differences between groups.
    """


class UncorrelatableError(ValueError):
    """Warned when an appropriate hypotheses-test cannot be drawn
    for the specific data type.
    """


def cramers_v(df: DataFrame,
              column_a: str,
              column_b: str,
              how: str = None,
              bias_correction: bool = True) -> Tuple[float, float]:
    """Calculates Cramér's V between categorical variables.

    See `Cramér's V <https://en.wikipedia.org/wiki/Cram%C3%A9r%27s_V>`_ for further information.

    Parameters
    ----------
    df : pyspark.sql.DataFrame
        The dataframe.
    column_a, column_b : str
        Names of the columns to be analysed.
    how : str, optional
        Allows a statistic other than Pearson's chi-squared to be used. See
        `scipy.stats.chi2_contingency <https://docs.scipy.org/doc/scipy-0.15.1/reference/generated/scipy.stats.chi2_contingency.html#scipy-stats-chi2-contingency>`_
        for more details. Defaults to None.
    bias_correction : bool, optional
        Whether to apply bias correction on the Cramér's V calculation. See
        `Cramér's V#Bias correction <https://en.wikipedia.org/wiki/Cram%C3%A9r's_V#Bias_correction>`_
        for details.

    Returns
    -------
    v : float
        Cramér's V between columns ``column_a`` and ``column_b``.
    p : float
        The p-value of the test.

    Examples
    --------
    .. jupyter-execute:: /examples/analysis/correlation/cramer_v.py

    Warns
    -----
    ChiSquareSmallFrequencyWarning
        If any of the contingency table's entries is less than 5. See
        `Pearson's_chi-squared_test#problems <https://en.wikipedia.org/wiki/Pearson's_chi-squared_test#Problems>`_
        for more information.
    CategoricalCorrelationSingleClassWarning
        If any of the columns has only one class, in which case it makes no sense to calculate correlation and the
        returned value is (0.0, 0.0).

    """
    contingency = df.crosstab(column_a, column_b).toPandas()

    contingency = contingency.set_index(contingency.columns[0])
    if (contingency.values < 5).any():
        warnings.warn("Some of the observed frequencies are less than 5, the Chi-Squared test may not be valid. "
                      "See https://en.wikipedia.org/wiki/Pearson's_chi-squared_test#Problems for more information.",
                      ChiSquareSmallFrequencyWarning)

    chi2, p, _, expected = stats.chi2_contingency(contingency, lambda_=how)

    if (expected < 5).any() and how != 'log-likelihood':
        warnings.warn("Some of the expected frequencies are less than 5, a likelihood ratio-based test statistic "
                      "may be better fit for this sample. Consider setting how=\"likelihood\". "
                      "See https://en.wikipedia.org/wiki/Pearson's_chi-squared_test#Problems for more "
                      "information.",
                      ChiSquareSmallFrequencyWarning)
    n = df.count()
    k = contingency.shape[0]
    r = contingency.shape[1]

    if k == 1 or r == 1:
        warnings.warn("Categorical column has only one class. Correlation cannot be calculated.",
                      CategoricalCorrelationSingleClassWarning)
        return 0.0, 0.0

    phi2 = chi2 / n

    if bias_correction:
        phi2 = max(0, phi2 - (k - 1) * (r - 1) / (n - 1))
        k = k - (k - 1) ** 2 / (n - 1)
        r = r - (r - 1) ** 2 / (n - 1)

    v = sqrt(phi2 / (min(k, r) - 1))

    return v, p


def z_test(df: DataFrame, column_a: str, column_b: str) -> Tuple[float, float]:
    """Calculates z-score and p-value of the distributions of a numeric variable over a categorical variable.

    Parameters
    ----------
    df : pyspark.sql.DataFrame
        The dataframe.
    column_a, column_b : str
        The name of the columns. One must be discrete and the other must be numeric.

    Returns
    -------
    z_score : float
        The z-score calculated for the distributions of ``num_column`` when grouped by ``cat_column``.
    p_value : float
        The p-value of the test.

    """
    try:
        num_column = next(c for c in (column_a, column_b) if AU.is_numeric(df, c))
        cat_column = next(c for c in (column_a, column_b) if AU.is_class(df, c, cat_threshold=None))
    except StopIteration:
        raise ValueError("One of `column_a` and `column_b` must be a class and the other must be numeric.")

    sigma, mu = df.agg(F.stddev_pop(num_column), F.mean(num_column)).collect()[0]
    m = df.groupBy(cat_column).agg(F.mean(num_column).alias('_mean')).collect()[0]['_mean']

    se = sigma / sqrt(df.count())

    zscore = (m - mu) / se
    pvalue = stats.norm.sf(abs(zscore)) * 2

    return zscore, pvalue


def pearson(df: DataFrame, column_a: str, column_b: str) -> Tuple[float, float]:
    """Calculates a Pearson correlation coefficient and the p-value for
    linear relationship between two datasets.

    See `scipy.stats.pearsonr <https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.pearsonr.html>`.

    Parameters
    ----------
    df : pyspark.sql.DataFrame
        The dataframe.
    column_a, column_b : str
        Names of the columns to be analysed.

    Returns
    -------
    coefficient : float
        Pearson's correlation coefficient between the columns
    p_value : float
        The 2 tailed p-value of the test.

    Examples
    --------
    .. jupyter-execute:: /examples/analysis/correlation/pearson.py

    """
    series = np.array(df.select(column_a, column_b).collect()).astype(float)
    return stats.pearsonr(*series.T)


def _correlation_test(df: DataFrame,
                      column_a: str,
                      column_b: str,
                      cat_threshold: Union[int, float, None] = .1) \
        -> Tuple[str, Callable[[DataFrame, str, str], Tuple[float, float]]]:
    if AU.is_class(df, column_a, cat_threshold=cat_threshold):
        if AU.is_class(df, column_b, cat_threshold=cat_threshold):
            return 'cramers_v', cramers_v
        elif AU.is_numeric(df, column_b):
            return 'z_test', z_test
    elif AU.is_numeric(df, column_a):
        if AU.is_class(df, column_b, cat_threshold=cat_threshold):
            return 'z_test', z_test
        elif AU.is_numeric(df, column_b):
            return 'pearson', pearson
    raise UncorrelatableError(f'Columns `{column_a}` and `{column_b}` on dataframe are not correlatable')


def one_to_many(df: DataFrame,
                column: str,
                subset: Union[str, List[str]] = 'auto',
                cat_threshold: Union[int, float, None] = .1) -> pd.DataFrame:
    """Calculates correlation between a single column and every other column in a subset of the dataframe.

    There are essentially three types of correlation, each calculated using a different method:

    - **Numeric-Numeric:** Between two numeric quantities. This is calculated using
      `Pearson correlation coefficient <https://en.wikipedia.org/wiki/Pearson_correlation_coefficient>`_,
      and the statistic ranges from -1.0 to 1.0, where -1.0 means strongly *negatively* correlated, 1.0 means
      strongly *positively* correlated and 0.0 means not correlated at all.

    - **Categorical-Categorical:** Between two nominal quantities. This is calculated using
      `Cramér's V <https://en.wikipedia.org/wiki/Cram%C3%A9r%27s_V>`_, and the statistic ranges
      from 0.0 to 1.0, where 0.0 means no relationship and 1.0 means strong relationship.

    - **Categorical-Numeric:** Between a categorical and a nominal quantity. This is calculated by grouping data by
      the categorical column and applying a `Z-test <https://en.wikipedia.org/wiki/Z-test>`_ to determine whether
      the distribution of the numeric column is the same between all categories (in which case the columns are not
      correlated). This test does not yield a correlation measure, but a z-score, which measures distance between
      distributions in terms of population standard deviations of the numeric column.

    Parameters
    ----------
    df : pyspark.sql.DataFrame
        The dataframe
    column : str
        Column name
    subset : {'auto', 'categorical', 'numeric'} or list of str, optional
        Determines which columns to analyze for correlation with `column`.
    cat_threshold : int or float or None, optional
        Maximum proportion between unique values and total that is considered to be categorical. Default is 0.1.

    See Also
    --------
    dextra.dna.core.analysis.utils.columns : for further explanation on ``subset``.
    dextra.dna.core.analysis.utils.is_categorical : for a detailed explanation on ``cat_threshold``.

    Returns
    -------
    pandas.DataFrame
        A Pandas DataFrame with columns ``test``, ``statistic`` and ``pvalue`` and one row for each column in the
        specified subset.

    Raises
    ------
    ValueError
        If ``column`` is neither numeric nor categorical.

    See Also
    --------
    correlation_matrix : Calculate correlation of every column to every other.

    Examples
    --------
    .. jupyter-execute:: /examples/analysis/correlation/one_to_many.py

    """
    subset = AU.correlatable_columns(df, subset, cat_threshold=cat_threshold) - {column} if isinstance(subset,
                                                                                                       str) else subset
    corr = {}
    for other in subset:
        name, test = _correlation_test(df, column, other, cat_threshold=cat_threshold)
        corr[other] = (name, *test(df, column, other))
    return (pd.DataFrame(corr, index=['test', 'statistic', 'pvalue']).T
            .astype({'test': str, 'statistic': float, 'pvalue': float}))


def correlation_matrix(
        df: DataFrame,
        subset: Union[str, List[str]] = 'auto',
        cat_threshold: Union[int, float, None] = .1) -> Tuple[pd.DataFrame, pd.DataFrame]:
    """Calculates a correlation matrix for columns in ``subset``.

    Since z-score works on a different scale than pearson coefficient and cramér's v, it is not shown in the
    correlation matrix. To analyze numeric-categorical correlation, use ``z_test_matrix``.

    See Also
    --------
    z_test_matrix : Check for correlation between each categorical-numeric pair of columns on a dataframe.
    dextra.dna.core.analysis.utils.is_categorical : for a detailed explanation on ``cat_threshold``.

    Parameters
    ----------
    df : pyspark.sql.DataFrame
        The dataframe
    subset : {'auto', 'categorical', 'numeric'} or list of str, optional
        Determines which columns to analyze for correlation.
    cat_threshold : int or float or None, optional
        Maximum proportion between unique values and total that is considered to be categorical. Default is 0.1.

    See Also
    --------
    dextra.dna.core.analysis.utils.columns : for further explanation on ``subset``.

    Returns
    -------
    cm : pandas.DataFrame
        Correlation matrix as a pandas DataFrame.
    pvalues : pandas.DataFrame
        Matrix with p-values for the given correlation values.

    Examples
    --------
    .. jupyter-execute:: /examples/analysis/correlation/correlation_matrix.py

    """
    if isinstance(subset, str):
        subset = list(AU.correlatable_columns(df, subset, cat_threshold=cat_threshold))

    subset = [c for c in subset if AU.is_correlatable(df, c, cat_threshold=cat_threshold)]

    cm = pd.DataFrame(index=subset, columns=subset, dtype=float)
    pvalues = pd.DataFrame(index=subset, columns=subset, dtype=float)
    for i, column in enumerate(subset):
        for _, row in one_to_many(df, column, subset=subset[i:], cat_threshold=cat_threshold).iterrows():
            if row.test != 'z_test':
                cm.loc[column][row.name] = cm.loc[row.name][column] = row.statistic
            pvalues.loc[column][row.name] = pvalues.loc[row.name][column] = row.pvalue
    return cm, pvalues


def z_test_matrix(df: DataFrame,
                  alpha: float = 0.05,
                  subset: Union[str, List[str]] = 'auto',
                  cat_threshold: Union[int, float, None] = .1):
    """Calculates a matrix of 0's and 1's indicating columns that are probably related in ``subset``.

    Parameters
    ----------
    df : pyspark.sql.DataFrame
        The dataframe.
    alpha : float
        Significance level (default: 0.05).
    subset : {'auto', 'categorical', 'numeric'} or list of str, optional
        Determines which columns to analyze for correlation.
    cat_threshold : int or float or None, optional
        Maximum proportion between unique values and total that is considered to be categorical. Default is 0.1.

    See Also
    --------
    dextra.dna.core.analysis.utils.columns : for further explanation on ``subset``.
    dextra.dna.core.analysis.utils.is_categorical : for a detailed explanation on ``cat_threshold``.

    Returns
    -------
    cm : pandas.DataFrame
        Boolean matrix as a pandas DataFrame. True values indicate a significance > 1 - alpha.
    pvalues : pandas.DataFrame
        Matrix with p-values for the given correlation values.

    Examples
    --------
    .. jupyter-execute:: /examples/analysis/correlation/z_test_matrix.py

    """
    subset = (list(AU.correlatable_columns(df, subset, cat_threshold=cat_threshold))
              if isinstance(subset, str)
              else subset)

    clazz = AU.correlatable_columns(df.select(subset), subset='class', cat_threshold=cat_threshold)
    numeric = AU.correlatable_columns(df.select(subset), subset='numeric')

    cm = pd.DataFrame(index=clazz, columns=numeric)
    pvalues = pd.DataFrame(index=clazz, columns=numeric)

    for i, column in enumerate(clazz):
        for _, row in one_to_many(df, column, subset=numeric, cat_threshold=cat_threshold).iterrows():
            cm.loc[column][row.name] = row.pvalue < alpha
            pvalues.loc[column][row.name] = row.pvalue
    return cm, pvalues



from typing import Union, Optional, List

import numpy as np
import pandas as pd
from pyspark.ml.pipeline import PipelineModel
from sklearn import cluster
from sklearn.base import BaseEstimator
from sklearn.decomposition import TruncatedSVD
from sklearn.linear_model import LogisticRegression
from sklearn.model_selection import GridSearchCV, BaseCrossValidator, RepeatedKFold, cross_val_score
from sklearn.pipeline import make_pipeline, Pipeline as SKLPipeline
from sklearn.preprocessing import Normalizer

SUPPORTED_ESTIMATORS = (BaseEstimator,)
SUPPORTED_CLUSTERS = ('kmeans', 'optics')


def logistic_regressor_gs(params: dict = None,
                          grid_searching: bool = True,
                          scoring: Optional[str] = None,
                          cv: Union[int, BaseCrossValidator] = 3,
                          workers: int = 1):
    model = LogisticRegression(solver='lbfgs', multi_class='auto', class_weight='balanced', max_iter=400)
    params = params or grid_searching and {'C': np.logspace(-3, 2, 10)}

    return (GridSearchCV(model, params, cv=cv, n_jobs=workers, scoring=scoring)
            if grid_searching
            else model)


def score_the_shit_out_of(model, x, y,
                          cv: Union[int, BaseCrossValidator] = None,
                          scoring: Optional[str] = None,
                          workers: int = 1,
                          verbose: int = 0):
    if not isinstance(model, SUPPORTED_ESTIMATORS):
        raise ValueError(f'Model {model} of class {model.__class__} is not supported. '
                         f'Use one of the following: {SUPPORTED_ESTIMATORS}.')

    if isinstance(model, BaseEstimator):
        # 5-2-fold cross-validation over the entire dataset.
        return (cross_val_score(model, x, y,
                                cv=cv or RepeatedKFold(n_splits=2, n_repeats=5),
                                scoring=scoring,
                                n_jobs=workers,
                                verbose=verbose)
                .mean())


def explain(model,
            feature_names: Union[PipelineModel, List[str]],
            features_used: int = 16,
            **kwargs):
    """Generate a report that explains the model's behavior.

    :param model:         the model that should be explained.
    :param feature_names: a list of feature names.
    :param features_used: the number of features shown during the explanation.
    :param \**kwargs:  see below.

    :Keyword Arguments:
        * *svd* -- SVD model used for unsupervised learning with KMeans

    :return: str, a report of the model
    """
    report = ''
    estimator = model.best_estimator_ if isinstance(model, GridSearchCV) else model

    if hasattr(estimator, 'feature_importances_'):
        g = estimator.feature_importances_
        most_important = np.argsort(g)[::-1][:features_used]

        report += f'\n### Most important features:\n'
        report += (pd.DataFrame([[feature_names[i], g[i]] for i in most_important],
                                columns=['term', 'weight'])
                   .set_index('term').T.to_string())
        report += '\n'

    elif isinstance(estimator, LogisticRegression):
        is_binary = len(estimator.coef_) == 1

        if is_binary:
            g = np.vstack((-estimator.coef_, estimator.coef_))
        else:
            g = estimator.coef_

        for label in estimator.classes_:
            c, = np.where(estimator.classes_ == label)[0]
            most_important = np.argsort(np.abs(g[c]))[::-1][:features_used]

            report += f'\n### Most contributing factors for the classification of label "{label}":\n'
            report += (pd.DataFrame([[feature_names[i], g[c, i]] for i in most_important],
                                    columns=['term', 'weight'])
                       .set_index('term').T.to_string())
            report += '\n'
    elif isinstance(estimator, cluster.KMeans):
        report += "Top terms per cluster:\n"

        if 'svd' in kwargs:
            svd = kwargs['svd']
            original_space_centroids = svd.inverse_transform(model.cluster_centers_)
            order_centroids = original_space_centroids.argsort()[:, ::-1]
        else:
            order_centroids = model.cluster_centers_.argsort()[:, ::-1]

        for i in range(model.cluster_centers_.size):
            report += "Cluster %d:" % i
            for ind in order_centroids[i, :features_used]:
                report += ' %s' % feature_names[ind]
            report += '\n'
    else:
        raise ValueError(f'We don\'t know how to explain the model {type(model)} yet. '
                         f'Try logistic regression machines or kmeans model.')

    return report


def _calculate_wcss(x, min_clusters, max_clusters):
    return [
        (cluster
         .MiniBatchKMeans(n_clusters=n, init='k-means++', n_init=1)
         .fit(x).inertia_)
        for n in range(min_clusters, max_clusters)
    ]


def _optimal_number_of_clusters(wcss, min_clusters, max_clusters):
    x1, y1 = min_clusters, wcss[0]
    x2, y2 = max_clusters, wcss[-1]

    distances = []
    for i, inertia in enumerate(wcss):
        x0 = i + min_clusters
        y0 = inertia
        numerator = abs((y2 - y1) * x0 - (x2 - x1) * y0 + x2 * y1 - y2 * x1)
        denominator = ((y2 - y1) ** 2 + (x2 - x1) ** 2) ** (1 / 2)
        distances.append(numerator / denominator)

    return np.argmax(distances) + min_clusters


def latent_semantic_analysis(n_components):
    svd = TruncatedSVD(n_components)
    normalizer = Normalizer(copy=False)
    return make_pipeline(svd, normalizer)


def discover_structure(x,
                       method: str = 'optics',
                       min_clusters: int = 2,
                       max_clusters: int = 32,
                       workers: int = 1):
    if method not in SUPPORTED_CLUSTERS:
        raise ValueError(f'Unknown method {method} passed. '
                         f'Available options are: {SUPPORTED_CLUSTERS}')

    if method == 'optics':
        return cluster.OPTICS(n_jobs=workers)

    wcss = _calculate_wcss(x, min_clusters, max_clusters)
    num_clusters = _optimal_number_of_clusters(wcss, min_clusters, max_clusters)
    return cluster.MiniBatchKMeans(n_clusters=num_clusters,
                                   init='k-means++',
                                   n_init=1)


def get_svd(model: SKLPipeline) -> TruncatedSVD:
    model = next(s for s in model.named_steps.values() if isinstance(s, TruncatedSVD))

    if model is None:
        raise ValueError('This model does not seem to have a svd transform.')

    return model



from typing import Union

from .. import io


class NoPrivateKeyError(ValueError):
    pass


class Crypto:
    """Asymmetric encryption manager."""

    def __init__(self,
                 public_key: Union[str, bytes, 'RSAPublicKey'] = None,
                 private_key: Union[str, bytes, 'RSAPrivateKey'] = None):
        """
        If neither {public_key} nor {private_key} are passed, they'll be generated automatically.
        Decryption is only possible when {private_key} is passed.

        :param public_key: RSA public key in PEM format (default: None). Ignored when {privateKey} is not {None}.
        :param private_key: RSA private key in PEM format (default: None). If it is passed, {publicKey} will be derived
                           from this key.
        """
        if private_key is not None:
            self._init_with_private_key(private_key)
        elif public_key is not None:
            self._init_with_public_key(public_key)
        else:
            self._private_key = self.generate_private_key()
            self._public_key = self._private_key.public_key()

    def _init_with_private_key(self, private_key):
        from cryptography.hazmat.backends import default_backend
        from cryptography.hazmat.primitives import serialization
        from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey

        if isinstance(private_key, RSAPrivateKey):
            self._private_key = private_key
        else:
            # TODO: Permitir senha na `privateKey`?
            self._private_key = serialization.load_pem_private_key(
                private_key,
                backend=default_backend(),
                password=None)
        self._public_key = self._private_key.public_key()

    def _init_with_public_key(self, public_key):
        from cryptography.hazmat.backends import default_backend
        from cryptography.hazmat.primitives import serialization
        from cryptography.hazmat.primitives.asymmetric.rsa import RSAPublicKey

        if isinstance(public_key, RSAPublicKey):
            self._public_key = public_key
        else:
            self._public_key = serialization.load_pem_public_key(public_key, backend=default_backend())
        self._private_key = None

    @classmethod
    def generate_private_key(cls):
        from cryptography.hazmat.backends import default_backend
        from cryptography.hazmat.primitives.asymmetric import rsa

        return rsa.generate_private_key(
            public_exponent=65537,
            key_size=2048,
            backend=default_backend()
        )

    @property
    def _private_bytes(self):
        from cryptography.hazmat.primitives import serialization

        return self._private_key.private_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PrivateFormat.PKCS8,
            encryption_algorithm=serialization.NoEncryption()
        )

    @property
    def _public_bytes(self):
        from cryptography.hazmat.primitives import serialization

        return self._public_key.public_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PublicFormat.SubjectPublicKeyInfo
        )

    @classmethod
    def load(cls, public_file: str = None, private_file: str = None) -> 'Crypto':
        """Loads a key with some name and creates an instance of Crypto.

        :param public_file: public key PEM file name
        :param private_file: private key PEM file name
        :return: encrypt-only Crypto set up with the loaded public key.
        """
        if private_file:
            return Crypto(private_key=bytes(io.storage.read(private_file), encoding='utf8'))

        if public_file:
            return Crypto(public_key=bytes(io.storage.read(public_file), encoding='utf8'))

        raise ValueError(
            '`Crypto.load` needs at least one valid key file to be passed. '
            'Either point to a public or private key path when calling it.')

    def encrypt(self, data: Union[str, bytes, bytearray]) -> bytes:
        """Takes some data and outputs it as an encrypted byte sequence.

        :param data: data to be encrypted.
        :return: encrypted data.
        """
        from cryptography.hazmat.primitives import hashes
        from cryptography.hazmat.primitives.asymmetric import padding

        if isinstance(data, str):
            data = bytes(data, encoding='utf8')

        return self._public_key.encrypt(data,
                                        padding=padding.OAEP(
                                            mgf=padding.MGF1(algorithm=hashes.SHA256()),
                                            algorithm=hashes.SHA256(),
                                            label=None
                                        ))

    def decrypt(self, data: bytes) -> bytes:
        """Takes some encrypted data and outputs it decrypted as a byte sequence.

        :param data: encrypted data to be decrypted.
        :return: decrypted data.
        """
        from cryptography.hazmat.primitives import hashes
        from cryptography.hazmat.primitives.asymmetric import padding

        if self._private_key is None:
            raise NoPrivateKeyError('No private key was set. Decrypting is not possible.')
        return self._private_key.decrypt(data, padding=padding.OAEP(
            mgf=padding.MGF1(algorithm=hashes.SHA256()),
            algorithm=hashes.SHA256(),
            label=None
        ))

    def save(self, public_file, private_file=None):
        """Saves cryptographic keys to the appropriate bucket in PEM format.

        :param public_file: filename to which the public key will be saved
        :param private_file: filename to which the private key will be saved
        """
        if self._private_key is not None and private_file is None:
            raise ValueError('Crypto instance contains a private key. The `private_file`'
                             'parameter must be passed in order to store it.')

        io.storage.write(public_file, self._public_bytes)

        if self._private_key is not None:
            io.storage.write(private_file, self._private_bytes)

    def __getstate__(self):
        if self._private_key is None:
            return {'public': self._public_bytes}
        else:
            return {'private': self._private_bytes}

    def __setstate__(self, state):
        if 'public' in state:
            self._init_with_public_key(state['public'])
        else:
            self._init_with_private_key(state['private'])



import abc
from typing import Dict, Any

from pyspark.sql import DataFrame


class StreamAdapter(metaclass=abc.ABCMeta):
    """Abstract stream environment adapter."""

    def read(self,
             filename: str,
             **options: Dict[str, Any]) -> DataFrame:
        """Reads a file from a stream.

        Parameters
        ----------
        filename : str

        options : dict
            Used as options during the read operation.
            It will vary according to the type of source being read.
            If the data is already loaded in memory (pandas, numpy,
            DataFrame), these options are ignored.

        Returns
        -------
        pyspark.sql.DataFrame

        Raises
        ------
        FileNotFoundError
            If file does not exist in storage.

        """
        raise NotImplementedError

    def write(self,
              filename: str,
              df: DataFrame,
              **options: Dict[str, Any]):
        """Writes a spark dataframe into a stream.

        Parameters
        ----------
        filename : str
            the location of the stream, can be a file, topic, table etc
        df: DataFrame
            frame to be written into the stream
        options : dict
            Used as options during the write operation. It will
            vary according to the type of source being written.

        """
        raise NotImplementedError



"""Describe The Different Processor Classes.
"""

import abc
import logging
from typing import List, Optional, Union, Dict

from pyspark.sql import functions as F

from infra.core.forge.processors import security
from infra.core.forge.processors import io
from infra.core.forge.utils import configs
from infra.core.forge.utils import to_list, stopwatch, NamedEntityMixin

I = Union[str, F.DataFrame, Dict[str, F.DataFrame]]


class Processor(NamedEntityMixin, metaclass=abc.ABCMeta):
    """Base Processor Class for Spark DataFrames.

    Parameters
    ----------
    inputs: I
        input stream or dictionary of input streams
    outputs: str
        saving location for the DataFrame processed
    config: Config
        the configuration under which this processor should run

    Attributes
    ----------
    loaded: DataFrame or map of frames
        Input frame (or map of frames) that store the loaded
        frames referenced by the paths passed in :code:`inputs`.
    processed: DataFrame, optional
        :code:`pyspark.sql.DataFrame` containing the result of
        the operations applied by this processor.
        This is set internally and should only be passed when testing.
    """

    SAVING_OPTIONS = {'mode': 'overwrite'}
    """Options passed to the :func:`io.stream.write` function.
    Through this option, one can change important aspects of how the data
    is persisted, such as partitioning schema or writing mode.
    """

    def __init__(self,
                 inputs: I,
                 outputs: str,
                 config: configs.Config = None,
                 loaded: Optional[I] = None,
                 processed: Optional[F.DataFrame] = None,
                 watches: Dict[str, stopwatch] = None):
        self.inputs = inputs
        self.outputs = outputs

        self.loaded = loaded
        self.processed = processed
        self.watches = watches or {}
        self.config = config

    @property
    def spark(self):
        """Retrieve the current spark session.

        Returns
        -------
        pyspark.sql.sessions.SparkSession
            Alias for the spark session contained within
            this processor's :code:`config`
        """
        return self.config.spark

    def setup(self, config=None) -> 'Processor':
        """Setup the processor for utilization.

        All custom setup should be done by re-implementing this method.
        """
        if config:
            self.config = config
        return self

    def perform(self) -> 'Processor':
        """Load the input data, if necessary, and apply the operations defined
           in the :func:`call` method to them. Finally, attributes the result
           to the :code:`processed` property.
        """
        logging.debug(f'run proc {self.fullname()}')

        with stopwatch(mode='silent') as w:
            self.load()

        self.watches['load'] = w

        with stopwatch(mode='silent') as w:
            if len(self.loaded) == 1:
                # Single inputs are passed as positional arguments.
                self.processed = self.call(*self.loaded.values())
            else:
                self.processed = self.call(**self.loaded)
        
        self.watches['proc'] = w

        return self

    def __call__(self, *args, **kwargs):
        return self.call(*args, **kwargs)

    @abc.abstractmethod
    def call(self, *args, **kwargs) -> F.DataFrame:
        """Defer to actual logic for data stream processing.

        This method must be implemented by the user in order to have any effect.
        The input is automatically unpacked in this method, so its override
        should either:

        1. have exactly one positional argument if it
           will be fed with a single input
        2. have arguments that match exactly the input dict keys

        Returns
        -------
        DataFrame
            the processed data stream
        """
        raise NotImplementedError

    def describe(self,
                 mode: str = 'simple') -> 'Processor':
        """Describe this processor with respect to its inputs and output.

        Parameters
        ----------
        mode: str
            the mode in which the description happens. Options are
            'simple' or 'explain'
        """

        describe_t = (self.config.logging.processors.describe
                      if 'describe' in self.config.logging.processors
                      else '%(processor)s')
        profile_t = (self.config.logging.processors.profile
                     if 'profile' in self.config.logging.processors
                     else 'load, proc and save time: %(load_t)s %(proc_t)s %(save_t)s')
        describe_c = {'processor': self.fullname(), 'inputs': self.inputs,
                      'proc': self.processed or '?', 'outputs': self.outputs}
        profile_c = {'load_t': '?', 'proc_t': '?', 'save_t': '?',
                     'started_at': (self.watches['load'].started_at
                                    if 'load' in self.watches
                                    else '?')}
        profile_c.update({f'{k}_t': round(v.elapsed, 3) for k, v in self.watches.items()})

        print(describe_t % describe_c, '',
              profile_t % profile_c,
              sep='\n', end='\n\n')

        if mode == 'explain' and self.processed:
            self.processed.explain()

        return self

    def load(self) -> 'Processor':
        """Load a processor's input into its space.

        :code:`Dict[str, DataFrame]` and :code:`DataFrame` are
        normalized into a single dictionary representation.
        """
        inputs = (self.inputs
                if isinstance(self.inputs, dict)
                else {'inputs': self.inputs})
        self.loaded = {k: io.stream.read(v)
                    for k, v in inputs.items()}

        return self

    def save(self) -> 'Processor':
        """Save the processed data into the path described by :code:`outputs`.
        """
        self._ensure_processed()
        
        with stopwatch(f'{self.fullname()} saving', mode='silent') as w:
            io.stream.write(self.processed,
                            self.outputs,
                            **self.SAVING_OPTIONS)
        
        self.watches['save'] = w

        return self

    def teardown(self) -> 'Processor':
        """Tear down processor after it's ran and its result is saved.

        By default, it does not do anything. The user must override this
        method and implement the appropriate logic when convenient.
        """
        return self

    def _ensure_processed(self):
        if self.processed is None:
            raise ValueError('The action you are attempting to perform '
                             'requires the data to be already processed '
                             'by this processor. Call the method '
                             '`Processor#perform` beforehand.')


class Rawing(Processor, metaclass=abc.ABCMeta):
    """Processor concerned with consolidating transient user data
    into a raw persisting bucket.

    Usually associated with the transient → raw transition.

    This processor passes :code:`mode='append'` to the spark writer when
    saving its results, which indicates input data will be accumulated.

    """

    SAVING_OPTIONS = {'mode': 'append'}

    def hash_sensible_info(self,
                           x: F.DataFrame,
                           columns: List[str]) -> F.DataFrame:
        """Hash data using an irreversible injective function.

        Parameters
        ----------
        x: DataFrame
            frame containing sensitive information
        columns: column or list of columns
            list of columns that could potentially be used to identify users

        Returns
        -------
        DataFrame
            A DataFrame without identifying features.
        """
        bits = self.config.security.encryption.bits
        length = self.config.security.encryption.length
        token = self.config.security.encryption.token

        for i in to_list(columns):
            x = x.withColumn(i, security.functions.anonymize(i, token, bits, length))

        return x


class Trusting(Processor, metaclass=abc.ABCMeta):
    """Processor concerned with validating data.

    Usually associated with the raw → trusted transition.

    """

    def discard_duplicates(self,
                           x: F.DataFrame,
                           id_field: Union[str, List[str]] = 'id',
                           date_field: str = None):
        """Erase duplicate records based on an identifying field.

        If {date_field} is passed, the records are sorted by most recent before
        dropping duplicates, prioritizing the discard of old entries.

        Parameters
        ----------
        x: DataFrame
           the frame possibly containing duplicates
        id_field: str, list
           column in the input frame containing the identifying data
        date_field: str, optional
           column in the input frame containing the date reference of each record

        Returns
        -------
        DataFrame
            A frame without duplicated rows.
        """
        if date_field:
            x = x.orderBy(F.desc(date_field))

        return x.dropDuplicates(to_list(id_field))


class Refining(Processor, metaclass=abc.ABCMeta):
    """Processor concerned with enriching consolidated data.

    Usually associated with the trusted → refined transition.

    """
    ML_DRIVER = Processor.models.drivers.Spark
    _ml_driver = None

    @property
    def ml(self):
        if not self._ml_driver:
            self._ml_driver = self.ML_DRIVER(
                config=self.config,
                weights_prefix=self.fullname().lower())
        return self._ml_driver


class Job(NamedEntityMixin):
    """Encapsulate a job's execution.

    Easily compose processors and their connection within a single object.

    Parameters
    ----------
    processors: list of Processor objects
        processors that must be sequentially executed in this job.
    inputs: I
        data stream that can be passed to the processors.
        This parameter is ignored if the user has not overridden the
        :func:`Job.build` method and passed it themselves.
    outputs: str
        path in which the job results are saved.
        This parameter is ignored if the user has not overridden the
        :func:`Job.build` method and passed it themselves.
    config: Config
        the configuration under which this job should run.
        It will be passed to the underlying processors during
        the setup stage.

    Examples
    --------
    .. jupyter-execute:: /examples/processors/pipeline.py
    """

    def __init__(self,
                 *processors: Processor,
                 inputs: Optional[I] = None,
                 outputs: Optional[str] = None,
                 config: configs.Config = None):
        self.processors = processors
        self.inputs = inputs
        self.outputs = outputs
        self.config = config

        self.staged = 0

    def setup(self, config: Optional[configs.Config] = None) -> 'Job':
        """Build and ensure all processors are within the same environment.

        Parameters
        ----------
        config: Config
                custom configuration in which this pipeline should run
        """
        logging.debug(f'setup {self.fullname()}')

        if config:
            self.config = config

        self.build()

        for p in self.processors:
            p.setup(config=self.config)

        return self

    def build(self):
        """Build the processing workflow, if necessary.

        By default, nothing is done and the processors passed during
        construction are used. However, one can overwrite this method to build
        more complicated associations between processors, as well as to
        encapsulate their links in order to perform integration testing.
        """
        logging.debug(f'build {self.fullname()}')

    def describe(self) -> 'Job':
        """Describe all processors within this pipeline.
        """
        print(type(self).__name__,
              f'  config: {self.config and self.config.env}',
              f'  staged: {self.staged}',
              f'  processors:',
              *(f'    - {p.fullname()}' for p in self.processors),
              sep='\n')

        return self

    def perform(self) -> 'Job':
        """Perform all operations within a pipeline and commit them.

        The methods :func:`Processor.perform`, :func:`Processor.describe` and
        :func:`Processor.save` are sequentially invoked for each processor
        within this pipeline.

        The :code:`stage` counter is increased at each committed step.
        """
        logging.debug(f'run job {self.fullname()}')

        for ix, p in enumerate(self.processors):
            p.perform().describe().save()
            self.staged = ix + 1

        return self

    def teardown(self) -> 'Job':
        """Teardown all processors within this pipeline by calling
        each individual :code:`teardown` method.
        """
        logging.debug(f'teardown {self.fullname()} ')

        for ix, p in enumerate(self.processors):
            p.teardown()

        return self


__all__ = ['Processor', 'Rawing', 'Trusting', 'Refining', 'Job']



from ink.core.forge.joins.text import testing, datasets

class SimpleIrisTest(testing.SparkTestCase):
    def setUp(self):
        self.iris = datasets.iris()

    def test_set_has_150_samples(self):
        expected = 150
        actual = self.iris.count()

        self.assertEqual(expected, actual)



import os

from infra.core.forge.utils.configs import Config, Test

HOME_DIR = os.path.expanduser('~')
CONFIG_DIR = os.getenv('CONFIG_DIR', os.path.join(HOME_DIR, 'dna'))
ENV = os.getenv('ENV', 'local')

config = Config(env=ENV, config_dir=CONFIG_DIR)

print(f'Using {config.env} environment.')

__all__ = ['Config', 'Test', 'config']



import abc
import re
import io
import os
from typing import List, Optional, Union, BinaryIO


class StorageAdapter(metaclass=abc.ABCMeta):
    """Abstract storage environment adapter."""

    @abc.abstractmethod
    def reader(self, filename: str) -> BinaryIO:
        """Opens a read stream for a file from storage.

        Parameters
        ----------
        filename : str

        Returns
        -------
        file-like object

        Raises
        ------
        FileNotFoundError
            If file does not exist in storage.

        """
        ...

    def read(self, filename: str) -> bytes:
        """Reads a file from storage.

        Parameters
        ----------
        filename : str

        Returns
        -------
        bytes

        Raises
        ------
        FileNotFoundError
            If file does not exist in storage.

        """
        with self.reader(filename) as file:
            return file.read()

    @abc.abstractmethod
    def writer(self, filename: str, append: bool = False) -> BinaryIO:
        """Opens a write stream for a file on storage.

        Parameters
        ----------
        filename : str
        append : bool, optional
            Whether to open file in append mode (default: False).

        Returns
        -------
        file-like object

        Raises
        ------
        FileNotFoundError
            If file does not exist in storage.

        """
        ...

    def write(self, filename: str, data: Union[bytes, BinaryIO], buffer_size: Optional[int] = io.DEFAULT_BUFFER_SIZE):
        """Writes data to a file on storage.

        Parameters
        ----------
        filename : str
        data : bytes or str
        buffer_size : int, optional
            Number of bytes to write at a time. Used only when data is a file-like object.

        """
        with self.writer(filename) as file:
            if not isinstance(data, (bytes, bytearray)):
                for chunk in iter(lambda: data.read(buffer_size), b''):
                    file.write(chunk)
            else:
                file.write(data)

    @abc.abstractmethod
    def isdir(self, path: str) -> bool:
        """Determines whether a given path refers to an existing directory.

        Parameters
        ----------
        path : str

        Returns
        -------
        bool

        """
        ...

    @abc.abstractmethod
    def isfile(self, path: str) -> bool:
        """Determines whether a given path refers to an existing file.

        Parameters
        ----------
        path : str

        Returns
        -------
        bool

        """
        ...

    def listdir(self, dirname: str, matching: Optional[str] = None) -> List[str]:
        """Lists contents of a directory on storage.

        Parameters
        ----------
        dirname : str
            Name of the directory.
        matching : str
            Regular expression to filter files with.

        Returns
        -------
        list of str
            A list with the names of the files in the directory.

        Raises
        ------
        FileNotFoundError
            If the specified directory does not exist.
        NotADirectoryError
            If the path does not refer to a directory.

        """
        files = self._listdir(dirname)
        if isinstance(matching, str):
            matching = re.compile(matching)
            return [f for f in files if not matching or matching.search(f)]
        else:
            return files

    @abc.abstractmethod
    def _listdir(self, dirname: str) -> List[str]:
        """Lists contents of a directory on storage.

        Parameters
        ----------
        dirname : str
            Name of the directory.

        Returns
        -------
        list of str
            A list with the names of the files in the directory.

        Raises
        ------
        FileNotFoundError
            If the specified directory does not exist.

        """
        ...

    @abc.abstractmethod
    def delete(self, filename: str):
        """Deletes a file from storage.

        Parameters
        ----------
        filename : str

        Raises
        ------
        FileNotFoundError
            If the specified directory does not exist.

        """
        ...

    @abc.abstractmethod
    def exists(self, filename: str) -> bool:
        """Checks whether a file exists in storage

        Parameters
        ----------
        filename : str

        Returns
        -------
        bool

        """
        ...

    @abc.abstractmethod
    def makedirs(self, filename: str):
        """Create a tree os directories if it does not yet exist.

        Parameters
        ----------
        filename : str
        """

    def _pre_copy(self, source: str, target: str, other: 'StorageAdapter'):
        if not self.exists(source):
            raise FileNotFoundError()

        (other or self).makedirs(target)

    def copy(self, source: str, target: str, other: Optional['StorageAdapter'] = None, move: bool = False):
        """Copies a file internally or between storages.

        Parameters
        ----------
        source, target : str
            Source and target file names.
        other : StorageAdapter, optional
            ``StorageAdapter`` for target storage environment.
        move : bool, optional
            Whether to delete the file after copying (default: False).

        Raises
        ------
        FileNotFoundError
            If source file does not exist.

        """
        if other is None:
            other = self

        self._pre_copy(source, target, other)

        if self.isfile(source):
            if other.isdir(target):
                target = os.path.join(target, os.path.basename(source))
            other.write(target, self.reader(source))

        elif self.isdir(source):
            if other.isfile(target):
                raise NotADirectoryError(f"Trying to copy a directory but target {target} is not a directory")
            for sub in self.listdir(source):
                self.copy(os.path.join(source, sub), os.path.join(target, sub))
        
        else:
            raise FileNotFoundError(f"{source} is not a file or a directory.")

        if move:
            self.delete(source)

    def move(self, source: str, target: str, other: Optional['StorageAdapter'] = None):
        """Moves a file internally or between storages.

        Parameters
        ----------
        source, target : str
            Source and target file names.
        other : StorageAdapter, optional
            ``StorageAdapter`` for target storage environment.

        Raises
        ------
        FileNotFoundError
            If source file does not exist.

        """
        self.copy(source, target, other, move=True)



import logging
from typing import List, Optional, Union

import numpy as np
import pandas as pd
from pyspark.sql import functions as F

from .stream_adapter import StreamAdapter
from infra.core.forge.utils.configs import spark


def _warn_unused_filename(filename):
    if filename is not None:
        logging.warning(
            f'Filename value "{filename}" passed to `stream#write` method '
            f'is ignored, as the distributed dataframe is being simply '
            f'converted into a pandas DataFrame.')


class MemoryStreamAdapter(StreamAdapter):
    def read(self,
             filename: Union[str, F.DataFrame, pd.DataFrame, np.ndarray],
             columns: List[str] = None) -> F.DataFrame:
        if isinstance(filename, F.DataFrame):
            return filename

        if isinstance(filename, np.ndarray):
            if not columns:
                columns = [f'_c{i}' for i in range(filename.shape[1])]
            filename = pd.DataFrame(filename, columns=columns)

        if isinstance(filename, pd.DataFrame):
            return spark().createDataFrame(filename)

    def write(self,
              df: F.DataFrame,
              filename: str,
              mode: str = 'overwrite',
              coalesce: Optional[int] = None,
              partitions: List[str] = None):
        _warn_unused_filename(filename)
        return df.toPandas()



import os
from logging import warning
from typing import Dict, Union, Optional

import yaml
from pyspark.sql import SparkSession
from pyspark.sql.types import Row

from . import utils
import findspark


def spark(mode='default') -> SparkSession:
    """Retrieve current spark session.

    :param mode: str
        The session mode. Should be either "default" or "test".

    :return: SparkSession
    """
    if mode == 'default':
        return SparkSession.builder.getOrCreate()
    elif mode == 'test':
        return (SparkSession.builder
                .master('local[2]')
                .appName('my-local-testing-pyspark-context')
                .enableHiveSupport()
                .getOrCreate())
    elif mode == 'deltalake':
        return SparkSession.builder \
            .config("spark.jars.packages", "io.delta:delta-core_2.12:0.7.0") \
            .config("spark.sql.extensions", "io.delta.sql.DeltaSparkSessionExtension") \
            .config("spark.sql.catalog.spark_catalog", "org.apache.spark.sql.delta.catalog.DeltaCatalog") \
            .config("spark.databricks.delta.schema.autoMerge.enabled", "true") \
            .getOrCreate()
    else:
        raise ValueError(f'Illegal value "{mode}" mode parameter. '
                         'It should be either "default", "test" or "deltalake".')


class Config:
    """Encapsulates all config properties for an execution.

    Parameters
    ----------
    env: str, default=DEFAULT_ENV
        environment
    config_dir: str, optional
        path to config folder containing desired configuration.
        Whenever passed, the :func:`load` method will be invoked
        and all yaml config files within that folder will be loaded
        and added as attributes to the calling object.

    Examples
    --------
    .. jupyter-execute::

        from dextra.dna.core import Config
        config = Config('local')
        config.lakes.transient

    """

    DEFAULT_ENV = 'local'
    SPARK_MODE = 'default'
    CONFIGS_FOLDER = os.environ.get('CONFIGS_DIR', 'config')
    CONFIG_EXTENSIONS = ('.yml', '.yaml')

    lakes: Row
    logging: Row
    security: Row

    def __init__(self,
                 env: str = None,
                 config_dir: Optional[Union[str, bool]] = None,
                 spark_mode: str = None):
        self.env = (env
                    or os.environ.get('ENV')
                    or self.DEFAULT_ENV)
        self.config_dir = (config_dir
                           or os.environ.get('CONFIG_DIR')
                           or os.path.join(self.CONFIGS_FOLDER, self.env))
        if spark_mode:
            self.SPARK_MODE = spark_mode

        if self.config_dir:
            self.load(raise_errors=False)

    @property
    def spark(self) -> SparkSession:
        return spark(self.SPARK_MODE)

    def load(self, raise_errors: bool = True) -> 'Config':
        """Load configuration from a yaml file.

        Parameters
        ----------
        raise_errors: bool
            whether errors should be raised or ignored when
            reading the configuration items
        """

        if not os.path.exists(self.config_dir):
            m = f'Attempting to load config from "{self.config_dir}", but no configuration was found.'

            if raise_errors:
                raise ValueError(m)
            else:
                warning(m)

            return self

        for f in os.listdir(self.config_dir):
            if not self.is_config_file(f):
                continue

            with open(os.path.join(self.config_dir, f)) as fp:
                g = fp.read()

                try:
                    g = g.format_map(os.environ)
                except ValueError:
                    ...

                for k, v in yaml.safe_load(g).items():
                    setattr(self, k, utils.as_row(v))

        return self

    def is_config_file(self, f) -> bool:
        """Check if a file is a plausible configuration file.
        """
        _, ext = os.path.splitext(f)
        return ext in self.CONFIG_EXTENSIONS

    def __repr__(self):
        return (f'<{self.__module__}.{type(self).__name__} at {hex(id(self))}, '
                f'logging: {self.logging}, lakes: {self.lakes}>')


class Test(Config):
    """Test configuration commonly used during test executions.

    """
    DEFAULT_ENV = 'test'
    SPARK_MODE = 'test'



class BaseError(RuntimeError):
    """Models Base Error.

    Generic base error class, used for centrality purposes during capturing.
    This class should not be used directly, in opposite of one of its subclasses below.
    """


class Convergence(BaseError):
    """Converge Error.

    Raised whenever a model or training procedure has failed to converge.
    """


class DegradedScore(BaseError):
    """Degraded Score Error.

    Raised whenever a model's score degrades over a new dataset sample.
    It can be used to signal the necessity of a new training.
    """


class Training(BaseError):
    """Training Error.

    Raised whenever a model cannot be fit properly.
    """


class TrainingNotFound(Training):
    """Training Not Found Error.

    Raise whenever a model's is required, but it's trained weights cannot be found.
    """



import os
import logging
from argparse import ArgumentParser

import infra.core.forge.archetype as ARCHETYPE_INITIAL


def parse_args():
    p = ArgumentParser('NAME_SIMPLE job')
    p.add_argument('--inputs', default=os.path.join(ARCHETYPE_INITIAL.config.lakes.transient, 'NAME.parquet'))
    p.add_argument('--outputs', default=os.path.join(ARCHETYPE_INITIAL.config.lakes.raw, 'NAME.parquet'))

    return p.parse_args()


def run(inputs, outputs):
    (ARCHETYPE_INITIAL.processors.NAME_SIMPLE(inputs=inputs, outputs=outputs)
     .setup(ARCHETYPE_INITIAL.config)
     .perform()
     .describe()
     .save()
     .teardown())



if __name__ == '__main__':
    logging.basicConfig(**ARCHETYPE_INITIAL.config.logging.default.asDict())
    args = parse_args()

    run(args.inputs, args.outputs)



import argparse
import logging

from . import generators, backend, consts


DESCRIPTION = 'Core Infra projects management tool for NLP Configuration'
ENV_PARAMS = {'nargs': '?', 'default': 'local', 'help': 'environment configuration used'}
VERBOSITY_CHOICES = [v.lower() for v in logging._nameToLevel]


def core_parser():
    p = argparse.ArgumentParser(description=DESCRIPTION)
    p.add_argument('--env', default='local', help='environment configuration used')

    e = p.add_mutually_exclusive_group()
    e.add_argument('--verbosity', '-v', type=str, choices=VERBOSITY_CHOICES, default=None)
    e.add_argument('--quiet', '-q', action='store_true', default=False)

    s = p.add_subparsers()

    g = s.add_parser('create', help='create core artifacts')
    g.add_argument('artifact', help='artifact to be created', choices=generators.GENERATORS.keys())
    g.add_argument('name', help='name of the creating artifact')
    g.add_argument('-t', '--to', default=None, help='location of the artifact')
    g.add_argument('-f', '--force', action='store_true', help='overwrite when already exists')
    g.add_argument('--arch-version', default='master', help='archetype version used (tag or branch)')
    g.set_defaults(operation=lambda args, _: generators.adapter(args).create())

    u = s.add_parser('build', help='build a core project')
    u.add_argument('env', **ENV_PARAMS)
    u.set_defaults(operation=lambda args, _: backend.adapter(args.env).build())

    u = s.add_parser('clean', help='clean project build')
    u.add_argument('env', **ENV_PARAMS)
    u.set_defaults(operation=lambda args, _: backend.adapter(args.env).clean())

    u = s.add_parser('start', help='start docker or cluster')
    u.add_argument('env', **ENV_PARAMS)
    u.set_defaults(operation=lambda args, _: backend.adapter(args.env).start())

    u = s.add_parser('explore', help='explore a core environment with jupyter')
    u.add_argument('env', **ENV_PARAMS)
    u.set_defaults(operation=lambda args, _: backend.adapter(args.env).explore())

    u = s.add_parser('stop', help='stop docker or cluster')
    u.add_argument('env', **ENV_PARAMS)
    u.set_defaults(operation=lambda args, _: backend.adapter(args.env).stop())

    u = s.add_parser('run', help='run project job')
    u.add_argument('job', help='path to job')
    u.set_defaults(operation=lambda args, remaining: backend.adapter(args.env)
                   .run(args.job, remaining))

    u = s.add_parser('sync', help='sync local notebooks with the ones in a cluster')
    u.add_argument('artifact', help='artifact to be synced', choices=consts.ARTIFACTS)
    u.add_argument('source', help='env containing the source artifacts')
    u.add_argument('target', help='env in which the artifacts will be updated')
    u.set_defaults(operation=lambda args, _: backend.adapter(args.source)
                   .sync(args.artifact, args.target))

    u = s.add_parser('test', help='run tests')
    u.set_defaults(env='test',
                   operation=lambda args, remaining:
                       backend.adapter(args.env).test(remaining))

    p.set_defaults(operation=lambda args, _: p.print_help())

    return p


def main():
    args, unknown = core_parser().parse_known_args()

    level = (getattr(logging, args.verbosity.upper(), None) if args.verbosity else
             logging.ERROR if args.quiet else
             logging.WARNING)

    logging.basicConfig(format=consts.LOGGING_FORMAT, level=level)

    try:
        args.operation(args, unknown)

    except KeyboardInterrupt:
        logging.info('Interrupted')

    except RuntimeError as e:
        logging.error(e)
        exit(-1)


if __name__ == '__main__':
    main()



import abc
import logging
import os
from typing import Optional, Union, List

from pyspark.sql import functions as F

from ...analysis.utils import correlatable_columns
from infra.core.forge.utils.configs import Config
from ....utils.utils import valid_or_default, NamedEntityMixin

FS = Optional[Union[str, List[str]]]


def extract_features(x: F.DataFrame,
                     target: str,
                     features: FS):
    """Extract features (if missing) by analyzing the DataFrame.

    If :code:`features` is passed (not None), then this will work as the
    identity function. Otherwise, the columns that contain learnable patterns
    are identified through the :func:`*.analaysis.utils.correlatable_columns`
    function.

    Parameters
    ----------
    x: DataFrame
        the frame onto which the model will be applied
    target: str
        the name of the column containing the target data, if any
    features: str or list of strings
        a column of a list of columns containing features

    Returns
    -------
    str or list of strings
        :code:`features`, if passed. Otherwise, returns a list of column names
        that contain patterns (according to the analysis module)

    """
    if features is None:
        features = [c for c in correlatable_columns(x) if c != target]

    if not features:
        raise ValueError('Cannot infer features from columns. '
                         'Try specifying it manually.')

    return features


class Driver(NamedEntityMixin,
             metaclass=abc.ABCMeta):
    """Orchestrates Machine Learning Models.

    Can be used to train, apply and store machine learning models throughout
    data pipeline's executions.

    Parameters
    ----------
    config: Config
        the current configuration held by the project.
        Should be passed by the caller (usually a job or Processor).

    caching: bool
        whether models trained and loaded should be cached or not.
        This is an abstract directive. Subclasses of Driver are
        responsible for caching whenever possible.

    """

    CACHING: bool = True

    def __init__(self,
                 config: Config,
                 caching: bool = None,
                 weights_prefix: Optional[str] = ''):
        self.config = config
        self.caching = valid_or_default(caching, self.CACHING)
        self.weights_prefix = weights_prefix

        self.cached_models = {}

    @abc.abstractmethod
    def load(self, weights: str):
        """Load The Model From A Name Passed.

        It's the driver's job to save the model, as an driver
        abstracts which kind of model is being handled. Nevertheless,
        :code:`config.lakes.models` is probably a good place to look for it.

        Returns
        -------
        ML model
            The model object loaded.
        """

    @abc.abstractmethod
    def save(self, model, weights: str):
        """Save a Trained Model to a Location Inferred from The Weights Passed.
        :code:`config.lakes.models` is probably a good place to look for it.

        Parameters
        ----------
        model: ML Model
            the model being saved
        weights: str
            the path into which the model should be saved

        Returns
        -------
        The ML model object loaded.
        """

    @abc.abstractmethod
    def learn(self,
              x: F.DataFrame,
              features: FS = None,
              target: str = 'target',
              weights: str = None):
        """Learn To Estimate a Target Based on Features Passed.

        This method should probably also save the model once it's trained.

        Parameters
        ----------
        x: DataFrame
            contains the training data
        features: str
            columns containing the features for the model
        target: str
            column containing the variable being estimated
        weights: str
            identifier for persisting the model. Whenever not available
            (:code:`None`), this should be inferred from the target.
        """

    @abc.abstractmethod
    def evaluate(self,
                 x: F.DataFrame,
                 features: FS = None,
                 target: str = 'target',
                 weights: str = None):
        """Evaluate Metrics For a Trained Model Over a Validation Dataset.

        Parameters
        ----------
        x: DataFrame
            contains the validation data
        features: str or list of strings
            column containing the variable being estimated
        target: str
            column containing the variable being estimated
        weights: str
            identifier for persisting the model. Whenever not available
            (:code:`None`), this should be inferred from the target.
        """

    @abc.abstractmethod
    def infer(self,
              x: F.DataFrame,
              features: FS = None,
              target: str = 'target',
              weights: str = None,
              on_errors: str = 'raise') -> F.DataFrame:
        """Infer A Specific Target Value from Available Features and a Trained Model.

        Parameters
        ----------
        x: DataFrame
            contains the validation data
        features: str or list of strings
            column containing the variable being estimated
        target: str
            column containing the variable being estimated
        weights: str
            identifier for persisting the model. Whenever not available
            (:code:`None`), this should be inferred from the target.
        on_errors: str ('raise', 'ignore')
            whether errors should be raised or ignored
            when inferring. If this is the case, the returned DataFrame should contain
            only null values in the the target column.
            Common errors are: incompatible features, missing model.

        Returns
        -------
        DataFrame
            The input DataFrame :code:`x`, plus the column :code:`target`
            containing the inferred values.

        """

    def trained(self, weights: str):
        """Check If a Trained Model Exists.

        This method should be overridden to produce more efficient checking strategies
        whenever they are available.

        Returns
        -------
        bool
            True if it exists, False otherwise.
        """
        try:
            self.load(weights)
        except FileNotFoundError:
            return False

        return True

    def location_from_target(self, target: str) -> str:
        """Determine a model's location from the target column.

        We append the suffix :code:`.p` here in order to represent
        a :code:`pickle` file.

        Parameters
        ----------
        target : The target column in which the data is trained over.

        Returns
        -------
        str
            The full location of the model.
        """
        if self.weights_prefix:
            return os.path.join(self.config.lakes.models,
                                self.weights_prefix,
                                target)

        return os.path.join(self.config.lakes.models,
                            target)

    def cache(self, weights: str, model=None):
        """Cache a model into memory or retrieve a cached one.

        Useful before persisting a model, holding it into memory for
        subsequential calls to :func:`infer` or :func:`evaluate` methods.

        Parameters
        ----------
        weights: str
            the path used when persisting the model,
            which uniquely identifies a trained model
        model: spark or sklearn model, optional
            a trained model, when setting cache, otherwise None

        Returns
        -------
        ML Model or None
            A trained ML model that has been previously cached,
            or None if :code:`model` parameter was not passed.
        """
        if not model:  # get
            logging.debug(f'driver cache get {weights}')
            return self.cached_models.get(weights)

        if self.caching:
            logging.debug(f'driver cache set {weights} {model}')
            self.cached_models[weights] = model

    def uncache(self, weights: str):
        """Clear model from cache and remove it from memory.

        This method is useful whenever a model consumes a large amount of
        memory, and it will no longer be necessary for the pipeline's
        execution.

        Parameters
        ----------
        weights: str
            the path used when persisting the model,
            which uniquely identifies a trained model
        """
        logging.debug(f'driver uncache {weights}')
        model = self.cached_models.pop(weights, None)

        if model is not None:
            del model

        return self

    def _balanced(self,
                  x: F.DataFrame,
                  target: str = 'target',
                  unbalance_rate: float = 1.5):
        """Balance a dataframe according to a target column through
        random undersampling.

        Parameters
        ----------
        x: DataFrame
            the frame onto which the model will be applied
        target: str
            the name of the column containing the target data, if any
        unbalance_rate: float in [1, inf)
            the maximum balance rate of :code:`target` column after this
            method is applied. Lower numbers create fully-balanced datasets,
            but might result in an aggressive reduction of the original set.

        """
        samples = dict(x.groupBy(target).count().collect())
        min_freq = min(samples.values())
        rates = {c: min(unbalance_rate * min_freq / f, 1) for c, f in samples.items()}

        return x.sampleBy(target, rates)



import os

import shutil
from typing import Optional, BinaryIO

from .storage_adapter import StorageAdapter


class FileSystemStorageAdapter(StorageAdapter):
    def reader(self, filename: str) -> BinaryIO:
        return open(filename, 'rb')

    def writer(self, filename: str, append: bool = False) -> BinaryIO:
        return open(filename, 'ab' if append else 'wb')

    def makedirs(self, filename: str):
        dirname = os.path.dirname(filename.rstrip('/'))
        if dirname:
            os.makedirs(dirname, exist_ok=True)

    def write(self, filename: str, *args):
        self.makedirs(filename)
        super().write(filename, *args)

    _listdir = staticmethod(os.listdir)
    isdir = staticmethod(os.path.isdir)
    isfile = staticmethod(os.path.isfile)

    def delete(self, filename: str):
        if os.path.isdir(filename):
            shutil.rmtree(filename)
        else:
            os.remove(filename)

    def copy(self, source: str, target: str, other: Optional['StorageAdapter'] = None, move: bool = False):
        self._pre_copy(source, target, other)

        if not isinstance(other, FileSystemStorageAdapter):
            return super().copy(source, target, other, move=move)

        if move:
            return shutil.move(source, target)

        if os.path.isdir(source):
            return shutil.copytree(source, target)

        shutil.copy(source, target)

    def exists(self, filename: str) -> bool:
        return os.path.exists(filename)



from functools import reduce, partial
from typing import Union, List, Dict, Any, Optional

import numpy as np
import pandas as pd
from pyspark.sql import functions as F

from .conforming import conform
from ..base import adapter, split_protocol_path
from ...utils import to_list, unpack, is_list

S = Union[str, F.DataFrame]
S = Union[S, List[S]]


_ADAPTERS_AVAILABLE = {
    'mem': ('ink.core.forge.joins.core.io.stream.adapters.memory', 'MemoryStreamAdapter'),
    'file': ('ink.core.forge.joins.core.io.stream.adapters.file', 'FileStreamAdapter'),
    'bigquery': ('ink.core.forge.joins.core.io.stream.adapters.big_query', 'BigQueryStreamAdapter')
}
_ADAPTERS = {}

adapter = partial(adapter,
                  adapters=_ADAPTERS,
                  adapters_available=_ADAPTERS_AVAILABLE)


def read(sources: S,
         **options: Dict[str, Any]) -> F.DataFrame:
    """Read heterogeneous sources into spark frames.

    Paths, frames, lists, pandas and numpy arrays are supported
    as well and will be converted into dataframes.

    Parameters
    ----------
    sources: str, list-like of str
        Path or list of paths to sources.
    options: **kwargs
        Used as options during the read operation.
        It will vary according to the type of source being read.
        If the data is already loaded in memory (pandas, numpy, DataFrame),
        these options are ignored.

    Returns
    -------

    DataFrame
        The input stream read, encapsulated into a spark dataframe.

    Examples
    --------
      >>> C.io.stream.read(['gs://bucket/data.csv', 'file:///base.parquet'])
      [DataFrame, DataFrame]
      >>> C.io.stream.read('file:///e.json')
      DataFrame
      >>> C.io.stream.read(pd.read_csv(...))
      DataFrame

    """
    many = is_list(sources)
    sources = to_list(sources)
    loaded = []

    for s in sources:
        if isinstance(s, (F.DataFrame, pd.DataFrame, np.ndarray)):
            frame = adapter('mem').read(s, **options)
        else:
            protocol, filename = split_protocol_path(s)
            frame = adapter(protocol).read(s, **options)

        loaded.append(frame)

    return loaded if many else unpack(loaded)


def write(df: F.DataFrame,
          filename: str,
          **options: Dict[str, Any]):
    """Write a spark dataframe into a stream.

    Parameters
    ----------
    df: DataFrame
        frame that will be saved

    filename: str
        location for the stream, a file topic or table name

    options: **kwargs
        Used as options during the write operation.
        It will vary according to the type of source being written.

    Examples
    --------
    >>> C.io.stream.write(C.datasets.iris(), 'gs://bucket/iris.parquet')

    """
    protocol, _ = split_protocol_path(filename)
    return adapter(protocol).write(df, filename, **options)


def merge(dfs: List[F.DataFrame]) -> Optional[F.DataFrame]:
    """Merge multiple frames into a single one.
    """
    return (reduce(F.DataFrame.unionByName, dfs)
            if dfs
            else None)


__all__ = ['read', 'write', 'merge', 'conform']



import abc
import logging
import os
import shutil

import yaml

from . import consts
from .base import Runner, context, assert_is_core_project, safe_name, as_path, C


class Generator(Runner, metaclass=abc.ABCMeta):
    def __init__(self,
                 config: C,
                 force: bool):
        super().__init__(config)
        self.name = self.config['NAME']
        self.target = self.config['_TARGET']
        self.arch_version = self.config['ARCH_VERSION']
        self.force = force

        self.simple_name = os.path.basename(self.name)


class Project(Generator):
    """Archetype Adapter

    Generate a new project using archetype as base.
    """
    def __init__(self, config: C, force: bool):
        super().__init__(config, force)
        self.simple_name = safe_name(self.simple_name)

    def create(self):
        if os.path.exists(self.target):
            if not self.force:
                raise ValueError(f'Project "{self.target}" already exists.')

            shutil.rmtree(self.target)

        logging.info(f'Creating project `{self.simple_name}`')

        (self
         .clone()
         .replace_with_actual_name())

        return self

    def clone(self):
        logging.info(f'Cloning archetype from `{consts.ARCH_REPO}`')
        os.system(f'git clone -q {consts.ARCH_REPO} {self.target} --depth 1')
        shutil.rmtree(os.path.join(self.target, '.git'))

        return self

    def replace_with_actual_name(self):
        logging.info(f'Replacing `archetype` occurrences by `{safe_name(self.simple_name)}`')

        roots = []

        for root, dirs, files in os.walk(self.target):
            target_root = self._rename_arch(root)

            if root != target_root:
                logging.debug(f'  mk {target_root}')
                os.makedirs(target_root, exist_ok=True)

            for f in files:
                source, target = os.path.join(
                    root, f), os.path.join(target_root, f)

                logging.debug(f'  mk {target}')

                with open(source) as s:
                    content = s.read()

                with open(target, 'w') as t:
                    t.write(self._rename_arch(content))

                if source != target:
                    logging.debug(f'  rm {source}')
                    os.chmod(target, os.stat(source).st_mode)
                    os.remove(source)

            roots.append((root, target_root))

        logging.debug(f'Deleting dangling archetype folders')
        for original_root, target_root in roots:
            if original_root != target_root and os.path.exists(original_root):
                logging.debug(f'  rm {original_root}')
                shutil.rmtree(original_root)

        return self

    def _rename_arch(self, content):
        return content.replace('archetype', self.simple_name)


class TemplateGenerator(Generator):
    DIRS: str
    SUFFIX: str = '.py'
    TEMPLATE_NAME: str = None

    @property
    def template_name(self):
        return self.TEMPLATE_NAME or self.__class__.__name__.lower()

    def create(self):
        assert_is_core_project('.')

        dirs = self.DIRS.format(**self.config)
        path = as_path(self.name, dirs, self.SUFFIX)

        if os.path.exists(path) and not self.force:
            return logging.warning(f'{self.template_name} {self.name} already exists at `{path}`')

        logging.info(f'Creating job {self.name} at `{path}`')
        os.makedirs(os.path.dirname(path), exist_ok=True)

        t = self.load_template()

        for key, value in self.default_context():
            t = t.replace(key, value)

        with open(path, 'w') as f:
            logging.debug(f' mk {path}')
            f.write(t)

    def load_template(self):
        name, ext = self.template_name, self.SUFFIX

        name = name.replace(ext, '')
        file_path = os.path.join(consts.TEMPLATES_DIR, f'{name}{ext}')

        if not os.path.exists(file_path):
            raise FileNotFoundError(f'Cannot find template {name} at `{file_path}`.')

        with open(file_path) as f:
            return f.read()

    def default_context(self):
        return (('NAME_SIMPLE', self.simple_name),
                ('NAME_DOTTED', self.name.replace('/', '.')),
                ('ARCHETYPE_INITIAL', self.config['SERVICE'][:1].upper()),
                ('ARCHETYPE', self.config['SERVICE']),
                ('NAME', self.name))


class Notebook(TemplateGenerator):
    DIRS = 'notebooks'
    SUFFIX = '.ipynb'


class Job(TemplateGenerator):
    DIRS = 'jobs'


class JobProcessor(Job):
    ...


class JobFull(Job):
    ...


class Processor(TemplateGenerator):
    DIRS = 'ink/core/forge/{SERVICE}/processors'


class Test(TemplateGenerator):
    DIRS = 'tests/unit/{SERVICE}'
    SUFFIX = '_test.py'


GENERATORS = {
    'project': Project,
    'notebook': Notebook,
    'job': Job,
    'job-processor': JobProcessor,
    'job-full': JobFull,
    'processor': Processor,
    'test': Test,
}


def adapter(args, verbose: bool = True):
    global GENERATORS

    if verbose: logging.info(f'Using {args.env} environment.')

    c = {} if args.artifact == 'project' else context(args.env)

    c.update(ARTIFACT=args.artifact,
             NAME=args.name,
             ARCH_VERSION=args.arch_version,
             _TARGET=args.to or args.name,
             _CORE_REPO=consts.CORE_REPO,
             _ARCH_REPO=consts.ARCH_REPO)

    logging.debug(yaml.dump({'config': c}))
    return GENERATORS[args.artifact](c, force=args.force)



from pyspark.sql import DataFrame, functions as F

import infra.core.forge.processors.core as C


class Rawing(C.processors.Rawing):
    """NAME_SIMPLE Rawing Processor.

    transient → raw

    Adds ingestion timestamp to the transient data
    and append it to the raw data.

    """
    SAVING_OPTIONS = {'mode': 'append'}

    def call(self, x: DataFrame):
        x = x.withColumn('ingested_at', F.current_timestamp())

        return x


class Trusting(C.processors.Trusting):
    """NAME_SIMPLE Trusting Processor.

    raw → trusted

    Discard samples without :code:`ids` and remove
    entry duplicates by removing the older ones.

    """
    SAVING_OPTIONS = {'mode': 'overwrite'}

    def call(self, x: DataFrame):
        x = x.where(x.id.isNotNull())
        x = self.discard_duplicates(x, 'id', 'created_at')

        return x


class Refining(C.processors.Refining):
    """NAME_SIMPLE Refining Processor.

    trusted → refined

    Sort data according to its :code:`id` and
    :code:`created_at` to improve reading performance.

    """
    SAVING_OPTIONS = {'mode': 'overwrite'}

    def call(self, x: DataFrame):
        x = x.orderBy('id', 'created_at')

        return x



from typing import Union

import pandas as pd
from pyspark import Broadcast
from pyspark.sql import types as T, functions as F
from pyspark.sql.column import Column
from sklearn.base import BaseEstimator

from .utils import as_model, as_features


# region sample prediction functions

def predict_likeness(col: Union[str, Column],
                     model: BaseEstimator,
                     label: int = 1) -> Column:
    """Predict Likeness of a Given Class.

    Parameters
    ----------
    col :
        Column describing features fed to the model.
        This column is commonly named 'features'.
    model :
        A scikit-learn inference model.
    label :
        The label of interest, over which the probability
        is extracted. If `None` is passed, then the highest
        probability is returned.

    Returns
    -------
    pyspark.sql.udf
        The spark function representing the operation that applied to model
        to the frame's column.
    """
    return F.udf(lambda x: float(as_model(model).predict_proba(as_features(x))[0].max()
                                 if label is None
                                 else as_model(model).predict_proba(as_features(x))[0, label]),
                 T.FloatType())(col)


def predict(col: Union[str, Column], model: Union[BaseEstimator, Broadcast]) -> Column:
    """Apply the model#predict method over a column.

    Parameters
    ----------
    col :
        Column describing features fed to the model.
        This column is commonly named 'features'.
    model :
        A scikit-learn inference model.

    Returns
    -------
    pyspark.sql.udf
        The spark function representing the operation that applied to model
        to the frame's column.
    """
    return F.udf(lambda x: str(as_model(model).predict(as_features(x))[0]))(col)


def transform(col: Union[str, Column], model: BaseEstimator) -> Column:
    """Apply the model#transform method over a column.

    Parameters
    ----------
    col :
        Column describing features fed to the model.
        This column is commonly named 'features'.
    model :
        A scikit-learn inference model.

    Returns
    -------
    pyspark.sql.udf
        The spark function representing the operation that applied to model
        to the frame's column.
    """
    return F.udf(lambda x: str(as_model(model).transform(as_features(x))[0]))(col)


# endregion

# region batching

def predict_likeness_batch(cols, model, label=1) -> Column:
    def _predict_likeness_batch(*x):
        return pd.Series(as_model(model).predict_proba(pd.concat(x, axis=1)).max(axis=1)
                         if label is None else
                         as_model(model).predict_proba(pd.concat(x, axis=1))[:, label])

    return F.pandas_udf(_predict_likeness_batch, 'double')(*cols)


def predict_batch(cols, model) -> Column:
    return F.pandas_udf(lambda *x: pd.Series(as_model(model)
                                             .predict(pd.concat(x, axis=1))),
                        'string')(*cols)

# endregion



import io
from functools import partial
from typing import List
from typing import Optional, Union

from ..base import (adapter,
                    split_protocol_path,
                    without_protocol,
                    join_protocol_path,
                    handle_error)
from ... import security

_ADAPTERS_AVAILABLE = {
    'file': ('infra.core.forge.processors.io.storage.adapters.file_system', 'FileSystemStorageAdapter'),
    'gs': ('infra.core.forge.processors.io.storage.adapters.google_cloud', 'GoogleCloudStorageAdapter')
}
_ADAPTERS = {}

adapter = partial(adapter,
                  adapters=_ADAPTERS,
                  adapters_available=_ADAPTERS_AVAILABLE)


def open(filename: str, mode: str = 'rb', encoding='utf-8', on_error: str = 'raise'):
    """Open files from multiple sources.

    # TODO: Implement cryptography for sources.
    """
    if not 3 >= len(mode) >= 1 or len(mode) > 1 and mode[1] not in {'+', 'b'}:
        return handle_error('invalid mode: ' + mode, on_error, ValueError)

    both = len(mode) > 1 and mode[1] == '+'
    binary = mode[-1] == 'b'

    protocol, filename = split_protocol_path(filename)
    try:
        ap = adapter(protocol)
        mode = '+' if both else mode[0]

        if mode[0] == '+':
            stream = io.BufferedRWPair(ap.reader(filename), ap.writer(filename))
        elif mode[0] == 'r':
            stream = ap.reader(filename)
        elif mode[0] in ('w', 'a'):
            stream = ap.writer(filename, append=mode[0] == 'a')
        else:
            return handle_error('invalid mode: ' + mode, on_error, ValueError)

        if not binary:
            stream = io.TextIOWrapper(stream, encoding)

        return stream
    except FileNotFoundError as error:
        handle_error(error, on_error)


def read(filename: str, encoding='utf-8', crypto: security.Crypto = None,
         on_error: str = 'raise') -> str:
    """Reads a file from storage.

    Parameters
    ----------
    filename : str
    crypto : security.Crypto
        Cryptography agent to be used for decryption.
    encoding : str, optional
        Encoding to read the file into (default: ``'utf-8'``).
    on_error : {'raise', 'log'}, optional
        Behavior in case errors happen during the process (default: ``'raise'``).

    Returns
    -------
    str

    Raises
    ------
    FileNotFoundError
        If file does not exist in storage.

    """
    protocol, filename = split_protocol_path(filename)
    try:
        data = adapter(protocol).read(filename)
        if crypto:
            data = crypto.decrypt(data)
        return data.decode(encoding)
    except FileNotFoundError as error:
        handle_error(error, on_error)


def write(filename: str, data: Union[bytes, str], encoding='utf-8', crypto: security.Crypto = None):
    """Writes data to a file on storage.

    Parameters
    ----------
    filename : str
    data : bytes or str
    encoding : str, optional
        Encoding to use when ``data`` is of type ``str`` (default: ``'utf-8'``).
    crypto : security.Crypto
        Cryptography agent to be used for encryption.

    """
    protocol, filename = split_protocol_path(filename)
    if isinstance(data, str):
        data = data.encode(encoding)
    if crypto:
        data = crypto.encrypt(data)
    adapter(protocol).write(filename, data)


def delete(filename: str, on_error: str = 'raise'):
    """Deletes a file from storage.

    Parameters
    ----------
    filename : str
    on_error : {'raise', 'log'}, optional
        Behavior in case errors happen during the process (default: ``'raise'``).

    Raises
    ------
    FileNotFoundError
        If the specified directory does not exist.

    """
    protocol, filename = split_protocol_path(filename)
    try:
        adapter(protocol).delete(filename)
    except FileNotFoundError as error:
        handle_error(error, on_error)


def copy(source: str, target: str, on_error: str = 'raise', move: bool = False):
    """Copies a file internally or between storages.

    Parameters
    ----------
    source, target : str
        Source and target file names.
    on_error : {'raise', 'log'}, optional
        Behavior in case errors happen during the process (default: ``'raise'``).
    move : bool, optional
        Whether to delete the file after copying (default: False).

    Raises
    ------
    FileNotFoundError
        If source file does not exist.

    """
    source_protocol, source_filename = split_protocol_path(source)
    target_protocol, target_filename = split_protocol_path(target)
    try:
        adapter(source_protocol).copy(source_filename, target_filename,
                                      other=adapter(target_protocol), move=move)
    except FileNotFoundError as error:
        handle_error(error, on_error)


def move(source: str, target: str, on_error: str = 'raise'):
    """Moves a file internally or between storages.

    Parameters
    ----------
    source, target : str
        Source and target file names.
    on_error : {'raise', 'log'}, optional
        Behavior in case errors happen during the process (default: ``'raise'``).

    Raises
    ------
    FileNotFoundError
        If source file does not exist.

    """
    copy(source, target, on_error, move=True)


def listdir(dirname: str, matching: Optional[str] = None, on_error: str = 'raise') -> List[str]:
    """Lists contents of a directory on storage.

    Parameters
    ----------
    dirname : str
        Name of the directory.
    matching : str
        Regular expression to filter files with.
    on_error : {'raise', 'log'}, optional
        Behavior in case errors happen during the process (default: ``'raise'``).

    Returns
    -------
    list of str
        A list with the names of the files in the directory.

    Raises
    ------
    FileNotFoundError
        If the specified directory does not exist.

    """
    protocol, dirname = split_protocol_path(dirname)
    try:
        return adapter(protocol).listdir(dirname, matching)
    except FileNotFoundError as error:
        handle_error(error, on_error)


def isdir(path: str, on_error: str = 'raise') -> List[str]:
    """Determines whether a given path refers to an existing directory.

    Parameters
    ----------
    path : str
    on_error : {'raise', 'log'}, optional
        Behavior in case errors happen during the process (default: ``'raise'``).

    Returns
    -------
    bool

    """
    protocol, path = split_protocol_path(path)
    try:
        return adapter(protocol).isdir(path)
    except FileNotFoundError as error:
        handle_error(error, on_error)


def isfile(path: str, on_error: str = 'raise') -> List[str]:
    """Determines whether a given path refers to an existing file.

    Parameters
    ----------
    path : str
    on_error : {'raise', 'log'}, optional
        Behavior in case errors happen during the process (default: ``'raise'``).

    Returns
    -------
    bool

    """
    protocol, path = split_protocol_path(path)
    try:
        return adapter(protocol).isfile(path)
    except FileNotFoundError as error:
        handle_error(error, on_error)


def exists(filename: str) -> bool:
    """Checks whether a file exists in storage

    Parameters
    ----------
    filename : str

    Returns
    -------
    bool

    """
    protocol, filename = split_protocol_path(filename)
    return adapter(protocol).exists(filename)


__all__ = [
    'split_protocol_path', 'without_protocol', 'join_protocol_path',
    'adapter',
    'open', 'read', 'write', 'delete', 'copy', 'move',
    'listdir', 'isdir', 'isfile', 'exists',
]



"""This module contains data analysis and visualization routines.

"""
import logging
from typing import Union, List, Optional

import numpy as np
import pandas as pd
from pyspark.sql import DataFrame, functions as F, types as T

from infra.core.forge.processors.analysis.utils import correlatable_columns
from infra.core.forge.processors import models 


def _get_reducer(r):
    from sklearn import manifold, decomposition

    reducers = {
        'tsne': manifold.TSNE,
        'pca': decomposition.PCA,
    }

    return (reducers[r.lower()]
            if isinstance(r, str)
            else r)


def distribution(df: DataFrame,
                 label: Optional[str] = None,
                 subset: Union[str, List[str]] = 'numeric',
                 sampling: Union[float, int] = 4096,
                 reducer_method: str = 'tsne',
                 cluster_method: str = 'optics',
                 workers: int = 1,
                 **kwargs):
    import seaborn as sns

    cols = sorted(correlatable_columns(df, subset=subset))
    if label: cols += [label]
    df = df.select(*cols)

    count = df.count()

    if isinstance(sampling, int):
        sampling = sampling / count

    if sampling < 1:
        logging.warning(f'The frames contains {count} samples. Data will be sampled.')
        df = df.sample(sampling)

    x = np.asarray(df.collect())

    if label:
        x, y = x[:, :-1], x[:, -1]
    else:
        m = models.discover_structure(x, cluster_method, workers)
        y = m.fit_predict(x)

    r = _get_reducer(reducer_method)(n_components=2)
    z = r.fit_transform(x)

    return sns.scatterplot(x=z[:, 0], y=z[:, 1], hue=y, **kwargs)


def sankey(df: DataFrame,
           sequence_col: str = 'journey',
           backend: str = 'bokeh',
           **kwargs):
    """Plot a Sankey diagram given a DataFrame with journey data.

    Parameters
    ----------
    df : pyspark.sql.DataFrame
        The dataframe. Each row should represent a journey.
    sequence_col : str, optional
        Name of the column containing the sequence of events for an user.
    backend : str, optional
        Holoviews backend to be used. ``None`` when no backend should be loaded (default).
    \**kwargs
        Keyword arguments to be passed to the sankey diagram plotting library.

        See `holoviews.Sankey <https://holoviews.org/reference/elements/bokeh/Sankey.html>`_
        for further information on what arguments are accepted.

    See Also
    --------
    dextra.dna.core.analysis.utils.journey : for further explanation on how to
        create a frame of sequence of events.

    Examples
    --------
    .. jupyter-execute:: /examples/analysis/visualization/sankey.py

    """
    import holoviews as hv

    if backend:
        hv.extension(backend)

    seq_fn = F.udf(lambda seq: list(zip(range(len(seq)), seq, seq[1:])),
                   returnType=T.ArrayType(T.ArrayType(T.StringType())))

    from_to = (df.withColumn('_st', F.explode(seq_fn(sequence_col)))
               .groupBy('_st')
               .count())

    df_sankey = pd.DataFrame([(f"{int(r['_st'][0]) + 1} {r['_st'][1]}",
                               f"{int(r['_st'][0]) + 2} {r['_st'][2]}",
                               r['count']) for r in from_to.collect()],
                             columns=['source', 'target', 'value'])
    plot = hv.Sankey(df_sankey)
    return plot.opts(**kwargs)



#! /usr/bin/env python3
import os

from setuptools import setup, find_namespace_packages

PROJECT_DIR = os.path.dirname(__file__)

INFO = open(os.path.join(PROJECT_DIR, 'INFO')).readlines()
INFO = dict((l.strip().split('=') for l in INFO))

DEPENDENCIES = open(os.path.join(PROJECT_DIR, 'requirements.txt')).readlines()

setup(name='archetype-core-nlp',
      version=INFO['version'],
      author=INFO['author'],
      author_email=INFO['author_email'],
      url=INFO['url'],
      python_requires='>=3.8',
      entry_points={
          'console_scripts': ['core=infra.core.forge:main']
      },
      packages=find_namespace_packages(include=['infra.core.forge','infra.core.forge.*']),
      namespace_packages=['infra', 'infra.core'],
      install_requires=[d for d in DEPENDENCIES if '://' not in d],
      package_data={'infra.core.forge': ['templates/*']},
      zip_safe=False)



"""Stream conformity functions.

Declares our main :code:`conform` function and its usual conformity
sub-functions. These are useful when trying to conform inconsistent
input user data.

"""
import re
from typing import Union, List, Optional, Callable

from pyspark.sql import functions as F, DataFrame

from ....utils.utils import to_list, valid_or_default

O = Union[Callable, str]
O = Optional[Union[O, List[O]]]


# region conforming functions

def sort_cols(df: DataFrame) -> DataFrame:
    """Sort the columns in a spark DataFrame to increase consistency.

    :param df: the non-normalized DataFrame being read.
    :return: The frame with its columns sorted in alphabetic order.

    Examples
    --------
    .. jupyter-execute:: /examples/io/stream/sort_cols.py
    """
    return df.select(*sorted(df.columns))


def lower_cols(df: DataFrame) -> DataFrame:
    """Lower-case all columns in the first-level of a frame.

    :param df: the processing frame.
    :return: A frame in which all columns are lower-case.

    Examples
    --------
    .. jupyter-execute:: /examples/io/stream/lower_cols.py
    """
    cols = [c.lower() for c in df.columns]
    assert_same_number_of_cols(df.columns, cols, 'lower_cols')

    return df.select([F.col(c).alias(r) for c, r in zip(df.columns, cols)])


def simple_cols(df: DataFrame) -> DataFrame:
    """Simplify the columns in the first-level of a frame.

    All non-alphanumeric characters are replaced by "_" in this operation.
    Any leading/trailing underscores are removed.

    :param df: the processing frame.
    :return: A frame with simple column names.

    Examples
    --------
    .. jupyter-execute:: /examples/io/stream/lower_cols.py
    """
    r = re.compile('[^0-9a-zA-Z]+')
    cols = [r.sub('_', c).strip('_') for c in df.columns]
    assert_same_number_of_cols(df.columns, cols, 'simple_cols')

    return df.select([F.col(c).alias(r) for c, r in zip(df.columns, cols)])


# endregion


# region validators

def assert_same_number_of_cols(a: List[str],
                               b: List[str],
                               op_name: str):
    """Assert the number of columns is the same.

    Useful to make sure the frame was not tempered with by
    the application of a possibly-destructive operation.

    :param a: the first (old) collection of columns.
    :param b: the second (new) collection of columns.
    :param op_name: The name of the operation being performed.

    Examples
    --------
    .. jupyter-execute:: /examples/io/stream/assert_same_number_of_cols.py
        :raises: ValueError
    """
    a, b = set(a), set(b)

    if len(a) != len(b):
        raise ValueError(f'Illegal operation `{op_name}` performed, as the '
                         f'number of columns in the frame would change from '
                         f'{len(a)} to {len(b)}. Difference:\n'
                         f'  A - B: {a - b}\n'
                         f'  B - A: {b - a}.')


# endregion


CONFORMING_OPS = {
    'sort_cols': sort_cols,
    'simple_cols': simple_cols,
    'lower_cols': lower_cols,
}
"""Dict: Map to all available conforming functions."""

DEFAULT_OPS = ('lower_cols', 'simple_cols', 'sort_cols')
"""Tuple[str]: Tuple of default operations executed in :code:`conform`."""


def adapter(op: Union[str, Callable]) -> Callable:
    """Retrieve a known conforming operation if its name is passed.
    Otherwise, this will work as the identity function.

    :param op: the operation's name or function of interest.
    :return: The operation referenced.
    """
    global CONFORMING_OPS

    if callable(op):
        return op

    if isinstance(op, str):
        if op not in CONFORMING_OPS:
            raise ValueError(
                f'Conforming operation `{op}` not found in `CONFORMING_OPS`. '
                f'Available options are: {CONFORMING_OPS.items()}')
        return CONFORMING_OPS[op]

    raise ValueError(f'Cannot infer an appropriate adapter for operation `{op}`. '
                     f'Valid arguments are callables or {list(CONFORMING_OPS.keys())}.')


def conform(dfs: Union[DataFrame, List[DataFrame]],
            ops: O = DEFAULT_OPS) -> List[DataFrame]:
    """Conform {DataFrames} according to some operation.

    :param dfs: frame or list of frames to be conformed.
    :param ops: str, function or list of str/functions
        Operation or list of operations used to conform the frames.
        Defaults to all conforming operations :code:`CONFORMING_OPS`.

    :return: The sames frames passed as arguments, but conformed
        according to the selected functions.

    Examples
    --------
    .. jupyter-execute:: /examples/io/stream/conform.py
    """
    if not dfs:
        return dfs

    dfs = to_list(dfs)
    ops = to_list(valid_or_default(ops, DEFAULT_OPS))

    for op in ops:
        op = adapter(op)
        dfs = [op(d) for d in dfs]

    return dfs



import logging

from py4j.protocol import Py4JJavaError
from pyspark import ml
from pyspark.sql import functions as F

from .base import FS, Driver, extract_features
from .. import errors
from ...analysis.utils import is_class
from ink.core.forge.joins.core.configs import Config
from ....utils.utils import is_list, valid_or_default


class Spark(Driver):
    """Orchestrates Spark Machine Learning Models.

    Parameters
    ----------

    config: Config
        the current configuration held by the project.
        Should be passed by the caller (usually a job or Processor).

    caching: bool
        whether models trained and loaded should be cached or not.
        This is an abstract directive. Subclasses of Driver are
        responsible for caching whenever possible.
        When not specified, defaults to :code:`Driver.CACHING`.

    scoring: str
        the metric being tracked during evaluation.
        When not specified, defaults to :code:`Spark.CACHING`.

    """
    SCORING: str = 'accuracy'
    TREES: int = 20
    KEEP_INTERMEDIATE: bool = False
    EVALUATOR = ml.evaluation.MulticlassClassificationEvaluator

    def __init__(self,
                 config: Config,
                 caching: bool = None,
                 scoring: str = None,
                 weights_prefix: str = '',
                 keep_intermediate: bool = None):
        super().__init__(config=config, caching=caching, weights_prefix=weights_prefix)

        self.scoring = valid_or_default(scoring, self.SCORING)
        self.keep_intermediate = valid_or_default(keep_intermediate, self.KEEP_INTERMEDIATE)

    def _indexing(self, x, target, target_is_class):
        if not target_is_class:
            return x, [], target

        si = ml.feature.StringIndexer(inputCol=target, outputCol=f'_{target}_indexed').fit(x)

        return x, [si], f'_{target}_indexed'

    def build(self,
              training: F.DataFrame,
              features: FS = None,
              target='target',
              **kwargs) -> ml.pipeline.Pipeline:
        logging.debug(f'driver build {self.fullname()}')

        target_is_class = is_class(training, target)

        training, stages, label_col = self._indexing(training, target, target_is_class)

        features = extract_features(training, target, features)
        logging.info('the following columns will be used as features:', features)

        if is_list(features):
            stages += [ml.feature.VectorAssembler(inputCols=features, outputCol='features')]

        stages += [
            ml.classification.RandomForestClassifier(
                featuresCol='features',
                labelCol=label_col,
                predictionCol=f'_{target}_prediction',
                probabilityCol=f'{target}_probability',
                numTrees=kwargs.get('trees', self.TREES)),
        ]

        if target_is_class:
            stages += [
                ml.feature.IndexToString(
                    inputCol=f'_{target}_prediction',
                    outputCol=f'{target}_prediction',
                    labels=stages[0].labels)
            ]

        return ml.pipeline.Pipeline(stages=stages)

    def load(self, weights: str) -> ml.pipeline.PipelineModel:
        logging.debug(f'driver load {weights}')
        return self.cache(weights) or self.load_from_storage(weights)

    def load_from_storage(self, weights: str) -> ml.pipeline.PipelineModel:
        """Load model from the storage backend.

        Parameters
        ----------
        weights: str
            the path into which the model should be saved

        Returns
        -------
        PipelineModel
            The loaded spark model
        """
        logging.debug(f'driver load model from storage {weights}')

        try:
            model = ml.pipeline.PipelineModel.load(weights)

        except Py4JJavaError as error:
            raise errors.TrainingNotFound(error)

        self.cache(weights, model)

        return model

    def save(self, model, weights) -> 'Spark':
        logging.debug(f'driver save {weights} {model}')
        self.cache(weights, model)

        model.write().overwrite().save(weights)

        return self

    def learn(self,
              x: F.DataFrame,
              features: FS = None,
              target: str = 'target',
              weights: str = None) -> 'Spark':
        logging.debug(f'driver learn {weights}')

        x = self._balanced(x, target)
        model = self.build(x, features, target).fit(x)

        location = self.location_from_target(target)
        self.save(model, weights or location)

        return self

    def evaluate(self,
                 x: F.DataFrame,
                 features: FS = None,
                 target: str = 'target',
                 weights: str = None,
                 balanced: bool = False):
        weights = weights or self.location_from_target(target)
        logging.debug(f'driver evaluate {weights}')

        model = self.load(weights)

        y = x if not balanced else self._balanced(x, target)
        y = model.transform(y)
        e = self.EVALUATOR(
            predictionCol=f'_{target}_prediction',
            labelCol=f'_{target}_indexed',
            metricName=self.scoring)

        return e.evaluate(y)

    def infer(self,
              x: F.DataFrame,
              features: FS = None,
              target: str = 'target',
              weights: str = None,
              on_errors: str = 'raise') -> F.DataFrame:
        weights = weights or self.location_from_target(target)
        logging.debug(f'driver infer {weights}')

        try:
            model = self.load(weights or self.location_from_target(target))

            y = (model
                 .transform(x)
                 .withColumnRenamed(model.stages[-1].getOutputCol(), target))

            y = self._cleanup(x, y, target)

            return y

        except errors.TrainingNotFound:
            if on_errors == 'raise': raise
            return x.withColumn(target, F.lit(None))

    def _cleanup(self, x, y, target):
        if self.keep_intermediate:
            return y

        intermediate = set(y.columns) - set(x.columns) - {target}
        return y.drop(*intermediate)

    def _warn_features_ignored_after_training(self, features):
        if features:
            logging.warning(
                f'The parameter `features` is ignored after the training of a '
                f'Spark model.  Value passed: `{features}`')



import os

ARCH_REPO = 'git@github.com:EneasJr-Rodrigues/docker-infra.git'
CORE_REPO = 'git@github.com:EneasJr-Rodrigues/docker-infra.git'
CORE_TEXT_REPO = 'git@github.com:EneasJr-Rodrigues/docker-infra.git'
CORE_IMAGE_REPO = 'git@github.com:EneasJr-Rodrigues/docker-infra.git'
CORE_SPEECH_REPO = 'git@github.com:EneasJr-Rodrigues/docker-infra.git'

DEPS = (
    ('arch_lakehouse_core', ARCH_REPO, 'CORE_VERSION'),
    ('arch_lakehouse_text', CORE_TEXT_REPO, 'CORE_TEXT_VERSION'),
    ('arch_lakehouse_image', CORE_IMAGE_REPO, 'CORE_IMAGE_VERSION'),
    ('arch_lakehouse_speech', CORE_SPEECH_REPO, 'CORE_SPEECH_VERSION'),    
)

LOGGING_FORMAT = '%(message)s'

TEMPLATES_DIR = os.path.join(os.path.dirname(__file__), 'templates')

ARTIFACTS = ('notebooks', 'jobs')



from . import errors, functions, drivers
from .base import *

__all__ = ['logistic_regressor_gs',
           'latent_semantic_analysis',
           'score_the_shit_out_of',
           'explain',
           'discover_structure',
           'functions',
           'errors',
           'drivers']
